'''
Sempy: A Python implementation of the spectral element method.
Copyright (C) 2011,  Stian Jensen.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
''' 

import sempy
import sempy.space.one_d.gmsh1D as mesh1d
import sempy.space.two_d.gmsh2D as mesh2d
import sempy.space.three_d.gmsh3D as mesh3d

import numpy as np

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from mpl_toolkits.mplot3d import axes3d
from matplotlib import cm
from matplotlib import rcParams
rcParams['text.usetex']=True
rcParams['text.latex.unicode']=True

import os,sys
import sempy.space.mesh_lib as mesh_lib

class Space:
    '''
    
    This class represents the implementation of the discrete space :math:`X_N`
    
    .. math::
    
       X_N\subset X & = H^1(\Omega)
       
       X_N &= \{v\in H^1(\Omega)\ |\ v|_{\Omega^k}\circ\mathcal F  \in\mathbb P_N(\hat\Omega),\ 
             k=0,\ldots, \\text{NOE}-1 \}
    
    where :math:`\Omega\in\mathbb R^d` is the computational domain and 
    :math:`\hat\Omega` is the reference domain.          
    An instance is created based on a Gmsh .msh file and 
    the chosen polynomial degree, :math:`N=n-1`. 
    
     
    
    :param filename: This is typically a Gmsh file with format extension 
                     *.msh*. However, :mod:`sempy` is shipped with a few 
                     ready to use meshes such as 
                     :literal:`line` for one dimensional problems, 
                     :literal:`square` for two dimensional problems and
                     :literal:`cube` for three dimensional problems.
    :type filename: string 
    :param n: Number of nodal points in each direction of an element. 
              Hence, the polynomial degree of the basis functions 
              is :math:`N=n-1`.
    :type n: int
    :param dim: Dimension :math:`d` of the computational domain 
                :math:`\Omega\in\mathbb R^d`. 
    :type dim: int
    
                     
    :attributes: * **noe** (*int*) - Number of spectral elements. 
                 * **noe_bc** (*int*) - Number of boundary elements. 
                 * **dof** (*int*) - Global degrees of freedom. 
                 * **points** (*array*) - GLL quadrature points 
                   :math:`\\underline{\\xi}\\in\\mathbb R^{N+1},\\ -1\\le \\xi_j\\le 1`.
                 * **weights** (*array*) - GLL quadrature weights 
                   :math:`\\underline{\\rho}\\in\\mathbb R^{N+1}` 
                   corresponding to the points :math:`\\underline{\\xi}`. 
                   Used to evaluate an integral over the reference domain, 
                   e.g.:
                 
                   .. math::
                      \\hat\\Omega\\in\\mathbb R^1: \\int_{\\hat\\Omega} f d\\hat\\Omega\\approx
                      \\sum_i\\rho_if_i
                      
                      \\hat\\Omega\\in\\mathbb R^2: \\int_{\\hat\\Omega} f d\\hat\\Omega\\approx
                      \\sum_i\\sum_j\\rho_i\\rho_jf_{ij}
                      
                 * **D** (*array*) - GLL derivative matrix 
                   :math:`\\underline D\\in\\mathbb R^{(N+1)\\times (N+1)}`. For a 1D problem:
    
                   .. math::
                      u_N'(\\xi_j)=\\sum_i l_i'(\\xi_j) u_i=\\sum_i D_{ji}u_i\\Rightarrow
                      D_{ji}=l_i'(\\xi_j)
                 
                 * **x,y,z** (*array*) - Coordinate arrays (given in local 
                   data distribution). 
                 * **jac** (*array*) - The Jacobian :math:`|\\underline{J}|`.
                 * **jac_s** (*array*) - The surface Jacobian 
                   :math:`|\\underline{J}^s|`. 
                 * **n_unit** (*array*) - Surface unit normal vector.
                 
                 
    **Example**::
    
       >>> import sempy
       >>> X_N = sempy.Space( filename = 'square', n = 3, dim = 2 )
       <BLANKLINE>
       **** Creating Space object     ****
       **** dim = 2
       **** noe = 84
       **** dof = 361
       ****   n = 3
       **** Finished creating instance.

       

             
    '''
    def __init__(self,filename, n, dim = 1):
        self.n = n
        self.dim = dim
        self.type = 'GLL'
        
        if dim == 1:
            if filename == 'line':
                path = os.path.dirname(mesh_lib.__file__)
                file_line = path+'/gmsh/line.msh'
                filename = file_line
            self._MeshD = mesh1d.MeshObject1D(filename,n,n_gmsh=3)
            self.x = self._MeshD.x 
            
        if dim == 2:
            if filename == 'square':
                path = os.path.dirname(mesh_lib.__file__)
                file_square = path+'/gmsh/square.msh'
                filename = file_square
            if filename == 'square_neumann':
                path = os.path.dirname(mesh_lib.__file__)
                file_square = path+'/gmsh/square_neumann.msh'
                filename = file_square
                #print '--filename=',filename

            self._MeshD = mesh2d.MeshObject2D(filename,n,n_gmsh=3)
            self.x, self.y=self._MeshD.x, self._MeshD.y
            self.z =  np.zeros((self._MeshD.noe, self.n,self.n),int)
            self.n_x = self._MeshD.n_x 
            self.n_y = self._MeshD.n_y
            
            self.n_unit  = np.zeros( (2,self._MeshD.noe_bc, self._MeshD.n),
                                      float)
            self.n_unit[0] = self._MeshD.n_x
            self.n_unit[1] = self._MeshD.n_y
        
        if dim == 3:
            if filename == 'cube':
                path = os.path.dirname(mesh_lib.__file__)
                file_square = path+'/gmsh/cube.msh'
                filename = file_square
            self._MeshD = mesh3d.MeshObject3D(filename,n)
            self.x, self.y, self.z = self._MeshD.x, self._MeshD.y,\
                                     self._MeshD.z
            
        self.physical = self._MeshD.physical
        
        self.noe = self._MeshD.noe
        self.noe_bc = self._MeshD.noe_bc
        self.dof = self._MeshD.dof
        self.weights = self._MeshD.weights
        self.points =self._MeshD.points
        self.D = self._MeshD.D
        self.jac = self._MeshD.jac
        
        self.jac_s = self._MeshD.jac_s
        self.theta = self._MeshD.theta
        self.theta_bc = self._MeshD.theta_bc
        self.bc_type = self._MeshD.bc_type
        
        self.fem = sempy.precond.SpaceFEM( self )
    
    def plot_mesh(self):
        '''
        Plots the computational mesh including type of boundary conditions. 
        Red circles indicate Dirichlet boundary conditions and green 
        circles indicate natural boundary conditions. 
        '''
        self._MeshD.plot_mesh()
        
    def plot_domain(self,font_size=20,show=True):
        '''Plot mesh.'''
        fig_number = np.random.randint(102,201)
        fig = plt.figure(fig_number) 
        if self.dim==2:
            thickness = 0.5
            n=self.n
            for k in range(self.noe):
                #for m in range(n):
                for i in range(n):
                    plt.plot(self.x[k,:,i],self.y[k,:,i],
                             color='k',linewidth=thickness)
                    #plt.plot(self.x[k,:,n-1],self.y[k,:,n-1],
                    #   color='k',linewidth=thickness)
                    plt.plot(self.x[k,i,:],self.y[k,i,:],
                             color='k',linewidth=thickness)
                    #plt.plot(self.x[k,n-1,:],self.y[k,n-1,:],
                    #  color='k',linewidth=thickness)
            plt.xlabel(r'$x$', fontsize = font_size )
            plt.ylabel(r'$y$', fontsize = font_size )
            plt.axis([self.x.min(),self.x.max(),self.y.min(),self.y.max()])
            plt.axes().set_aspect('equal')
            ax = plt.gca()
            for label in ax.xaxis.get_ticklabels():
                label.set_fontsize( font_size )
            for label in ax.yaxis.get_ticklabels():
                label.set_fontsize( font_size )
            if show:
                plt.show()
            
    def plot_scalar(self, u, n_int = 0):
        '''
        Plots a scalar :math:`u(x)`. 
        '''
        self._MeshD.plot_scalar(u, n_int)
        
    def plot_basis_function(self,i=0,show=True,font_size=20):
        '''
        Plots the :math:`i`'th basis function. 
        
        :param i: Global node number, :math:`i\in\{0,\ldots,\\text{DOF}-1\}`. 
        '''
        print 'show=',show
        self._MeshD.plot_basis_function(i,show=show,font_size=font_size)
        
    def mapping_q(self,v):
        '''
        The action of the mapping operator :math:`Q` on a vector :math:`v`.
        
        :param array v: Basis coefficients in global data representation.
        :returns: Basis coefficients in local data representation
        
        This method maps a vector of global ordering to a local data 
        representation. This operator is used in the assembly or the action 
        of a global matrix on a vector,
        
        .. math::
           \underline w=\underline  Q^T\underline A_L\underline Q\underline v=\underline A\,\underline v
           
        where :math:`\underline A_L` is e.g. the local Laplacian or the local mass matrix. 
        
        The matrix-vector product 
        :math:`\underline v_{local}=\underline Q\underline v_{global}` 
        can be achieved as follows::
           
           import sempy
           import numpy as np
           
           X = sempy.Space( filename = 'square', n = 5, dim = 2 )      

           u_global = np.ones( ( X.dof ) , float )
           u_local = X.mapping_q( u_global )
        
        '''
        return self._MeshD.mapping_q(v)
        
    def mapping_qt(self,u):
        '''
        The action of the transpose :math:`Q^T` of the mapping operator 
        :math:`Q`. 
        
        :param array u: Basis coefficients in local data representation.
        :returns: Basis coefficients in global data representation 
        
        This method maps an array of local ordering to a global data 
        representation. Also known as the gather operation. The matrix 
        vector product :math:`v_{global}=Q^Tv_{local}` is performed 
        by typing::
        
           import sempy
           import numpy as np

           X = sempy.Space( filename = 'square', n = 5, dim = 2 )      
    
           u_local = np.ones( ( X.noe, X.n, X.n ), float )
           u_global= X.mapping_qt( u_local )
        
        '''
        return self._MeshD.mapping_qt(u)
        
    def local_to_global(self,u):
        '''
        Maps the basis coefficients of a function 
        :math:`u\in H^1(\Omega)` from local to global 
        data representation.
        
        :param array u: Basis coefficients in local data representation.
        :returns: Basis coefficients in global data representation
        
        Usage::
           
           import sempy
           
           X = sempy.Space( filename = 'square', n = 5, dim = 2 )
           
           u = sempy.Function( X, basis_coeff =1.0 )
           
           u_global = X.local_to_global( u.basis_coeff )
           
        '''
        return self._MeshD.local_to_global(u)
    
    def mask(self,w):
        '''
        This method masks nodal point values associated with Dirichlet 
        boundary conditions.
        
        :param array w: Basis coefficients in global data representation 
                  subject to masking.
        :returns: Masked basis coefficients in global data representation.
        
        '''
        return self._MeshD.mask(w)
        
    def mask_matrix(self,A):
        '''
        This method masks entries in a matrix associated with Dirichlet 
        boundary conditions.
        
        :param A: Matrix subject to masking.
        :type A: PysparseMatrix
        :returns: The masked matrix.
        
        '''
        return self._MeshD.mask_matrix(A)
        
    def boundary_value(self,u):
        return self._MeshD.boundary_value(u)
        
    def l2_norm(self,u):
        '''
        Computes the :math:`L^2(\Omega)` - norm
           
        .. math::
           ||u||_{L^2(\\Omega)}=\\bigg(\\int_{\\Omega} u^2d\\Omega\\bigg)^{1/2}
              
        using GLL quadrature. 
           
        :param array u: Basis coefficients in global data structure, i.e. 
                        :math:`u\\in\\mathbb R^{dof}` .
        :returns: The :math:`L^2` - norm as a floating number.
                   
        Usage::
           
             import sempy
             
             X = sempy.Space( filename = 'square', n = 5, dim = 2 )      
             u = np.ones( ( X.dof ), float )
             print 'L2-norm=',X.l2_norm( u )
        '''
        return self._MeshD.l2_norm(u)
        
    def quadrature(self,f):
        '''
        Computes the value :math:`\\kappa` of the integral 
        
        .. math::
           \\kappa = \\int_\\Omega f\\,d\\Omega 
           
        for :math:`f\\in L^2(\\Omega)` with GLL quadrature rules. 
        
        :param array f: Basis coefficients of the function to be integrated 
                        given in local data representation 
                        (since :math:`f\\in L^2(\\Omega)`). 
        :returns: Value :math:`\\kappa` of the integral.
        
        Usage::
           
             import sempy
             X = sempy.Space( filename = 'square', n = 5, dim = 2 )
    
             u = sempy.Function( X, basis_coeff = X.x ) 
             kappa = X.quadrature( u.basis_coeff )
        '''
        return self._MeshD.quadrature(f)
    
    def theta_phys(self, *arguments):
        '''
        Local to global connectivity method. Takes as input the local node 
        description and returns the global node number and the physical name 
        of the element. 

        :arguments: Local node number. 
        :returns: * **i** (*int*)- Global node number. 
                  * **phys** (*int*) - Physical name.         
           
             
        A two dimensional example::
        
            import sempy
            
            X = sempy.Space( filename = 'square', n = 2, dim = 2 )
    
            for k in range( X.noe ):
                for m in range( X.n ):
                    for n in range( X.n ):
                        i, phys = X.theta_phys(k,m,n)
                        print '(k,m,n)=(',k,m,n,') i=',i
           
        
        '''
        return self._MeshD.theta_phys(*arguments)
        
    def theta_phys_bc(self, *arguments):
        '''
        Local to global connectivity method for boundary points. 
        
        .. Takes as input the local node 
           description and returns the global node number and the physical 
           name of the element. 
        
        
        :arguments: Local boundary node number. For 1D (k), for 2D (k,m) and for 3D (k,m,n). (k[,m[,n]])
        :returns: * **i** (*int*) - Global node number. 
                  * **phys** (*int*) - Physical name.
                  * **bc_type** (*string*) - Type of boundary condition,
                    e.g. "Dir" or "Neu", as specified in the Gmsh .geo file. 
                  * **element_edge** (*array*) - Holds the edge number 
                    and element number. 
        
        This routine can be used to apply boundary conditions. Note however 
        that the :class:`Function` class has easier ways of 
        dealing with this issue. 
        
        For example, for a 2D problem::
        
             import sempy
             
             X = sempy.Space( filename = 'square', n = 2, dim = 2 )
             u = sempy.Function( X, basis_coeff = 0.0 )
    
             u_global = u.glob()
             u_dirichlet = 10.0
             
             for k in range( X.noe_bc ): # Loops over the boundary elements.
                for m in range( X.n ):
                   [ i, phys, bc_type, edge ] = X.theta_phys_bc( k, m )
                   if bc_type=='Dir': # Assign BC value
                      u_global[i] = u_dirichlet
                
             u.basis_coeff = X.mapping_q( u_global )
             u.plot_wire()
                          
        '''
        return self._MeshD.theta_phys_bc(*arguments)
        
    def interpolation(self,u_N,m):
        '''
        Action of the interpolation operator :math:`I_N^M`
        
        .. math::
        
           u=I_N^M u_N
           
        where :math:`M=m-1`. 
        
        :param array u_N: Basis coefficients of the function 
                          to be interpolated.
        :type u_N: array 
        :param m: Number of nodal points in each elemental direction.
        :type m: int
        :returns: Basis coefficients of the interpolated function 
                  of polynomial degree :math:`M=m-1`. 
        
        Usage::
        
           import sempy
    
           X = sempy.Space( filename = 'square', n = 5, dim = 2 )
    
           u = sempy.Function( X, basis_coeff = X.x )
           u_int = X.interpolation( u.basis_coeff, 2 )
           u_new = X.interpolation( u_int, X.n )
           u.set_value( u_new )
           u.plot_wire()
        '''
        return self._MeshD.interpolation(u_N,m)
        
    def hyper_filter(self,u_tilde,alpha=0.25):
        '''
        A filter for hyperbolic problems (see Deville et al.). 
        
        .. math::
           u = F_\\alpha \\tilde u
           
        where the filter is
        
        .. math::
           F_\\alpha := \\alpha \\Pi_{N-1} + (1-\\alpha) I_N^N,\quad 
           \\Pi_{N-1} = I_{N-1}^N I_{N}^{N-1}
           
        Here, the weighting factor :math:`\\alpha` is typically in the 
        range :math:`0.05<\\alpha<0.3`. 
           
        :param array u_tilde: Basis coefficients of the function to 
                              be filtered. 
        :param float alpha: Weighting factor.
        :returns: Filtered basis coefficients. 
        '''
        return self._MeshD.hyper_filter(u_tilde,alpha)
        
    def gradient_vector(self,u):
        '''
        Computes the gradient vector of :math:`u`, i.e
        
        
        .. math::
        
           \\text{1D}:\quad \\nabla u= \partial u/\partial x
           
           \\text{2D}:\quad \\nabla u= (\partial u/\partial x,\partial u/\partial y)^T
           
           \\text{3D}:\quad \\nabla u= (\partial u/\partial x,\partial u/\partial y,\partial u/\partial z)^T
           
        :param u: The function to be differentiated
        :returns: The components of the gradient vector. 
        
        Usage::
          
          import sempy
    
          X = sempy.Space( filename = 'square', n = 5, dim = 2 )
    
          u = sempy.Function( X, basis_coeff = X.x * X.x )          
          u_x,u_y = X.gradient_vector(u.basis_coeff)
        
        '''
        return self._MeshD.gradient_vector( u )
        
    def surface_unit_normal(self):
        return self._MeshD.surface_unit_normal()
        
    def surface_flux(self,u):
        return self._MeshD.surface_flux(u)
 


def bc_function(Space,value=[0.0], bctype=[1]):
    g = np.zeros((Space.noe_bc, Space.n),float)
        
    if len(value) < len(bctype) or len(value) > len(bctype):
        print '---> Error in bc_function. len(value) does not equal len(bctype)'
        print '---> For example: value = [1.0, 2.0], bctype = [1,2]'
        #return 
    
    val    = np.zeros((len(value),Space.noe, Space.n, Space.n),float)
    val_bc = np.zeros((len(value),Space.noe_bc, Space.n),float)
    for j in range(len(value)):
        val[j][:,:,:] = value[j]
        val_bc[j] = Space.boundary_value(val[j])
        
    for j in range(len(value)):
        for k in range(Space.noe_bc):
            for m in range(Space.n):
                i, phys, bc_type, element_edge = Space.theta_phys_bc(k,m)
                if phys == bctype[j]:
                    g[k,m] = val_bc[j][k,m]
    return g
    
'''
class BoundaryFunction:
'''
    #BC.
'''
    def __init__(self,Space,value=[0.0], bctype=[1]):
        self.Space = Space
        self.coeff = np.zeros((self.Space.noe_bc, \
                               self.Space.n),float)
        
        if len(value) < len(bctype) or len(value) > len(bctype):
            print '---> Error in bc_function. len(value) does not equal len(bctype)'
            print '---> For example: value = [1.0, 2.0], bctype = [1,2]'
            #return 
    
        val    = np.zeros((len(value),self.Space.noe, \
                          self.Space.n, self.Space.n),float)
        val_bc = np.zeros((len(value),self.Space.noe_bc, \
                           self.Space.n),float)
        for j in range(len(value)):
            val[j][:,:,:] = value[j]
            val_bc[j] = self.Space.boundary_value(val[j])
        
        for j in range(len(value)):
            for k in range(self.Space.noe_bc):
                 for m in range(self.Space.n):
                    i, phys, bc_type, element_edge = self.Space.theta_phys_bc(k,m)
                    if phys == bctype[j]:
                        self.coeff[k,m] = val_bc[j][k,m]
                        #self.__setitem__[k,m] = val_bc[j][k,m]
                        #self.__setitem([k,m],val_bc[j][k,m])
        #return self.g
    #def __setitem__(self,key,value):
    #    self[key]=value
        
        
        
    def plot(self):
        
        x_bc = self.Space.boundary_value(self.Space.x)
        y_bc = self.Space.boundary_value(self.Space.y)
        u = np.zeros((self.Space.noe, self.Space.n,self.Space.n),float)
        
        for k in range(self.Space.noe_bc):
            for m in range(self.Space.n):
                [i, phys, bc_type, element_edge] = self.Space.theta_phys_bc(k,m)
                k_element = element_edge[0]
                if element_edge[1] == int(0):
                    u[k_element,m,0] = self.coeff[k,m]
                if element_edge[1] == int(1):
                    u[k_element,m,self.n-1] = self.coeff[k,m]
                if element_edge[1] == int(2):
                    u[k_element,0,m] = self.coeff[k,m]
                if element_edge[1] == int(3):
                    u[k_element,self.n-1,m] = self.coeff[k,m]
                    
        fig = plt.figure()
        ax = axes3d.Axes3D(fig)
        for k in range(self.Space.noe):
            ax.plot_wireframe(self.Space.x[k,:,:],self.Space.y[k,:,:],u[k,:,:])
        plt.show()
'''
                

class BoundaryFunction:
    '''
    Creates a function in the discrete space restricted to the boundary 
    :math:`\Gamma` of the computational domain :math:`\Omega`. 
    
    .. .. math::
       Y_N=\{v\in L^2(\Omega)\ |\ v|_{\Omega^k}\in\mathbb P_N(\Omega^k),\ 
             k=0,\ldots, \\text{NOE}-1 \}
    
    :param Space: The function space.
    :type Space: :class:`sempy.Space`
    :param array basis_coeff: Basis coefficients
    
    This class can be used in concert with the class 
    :class:`sempy.operators.Laplacian` to set Neumann and Robin 
    boundary conditions. It can be noted that the values of the *basis_coeff* 
    associated with Dirichlet conditions are masked. 
    
    **Example**::
    
       import sempy
      
       X = sempy.Space( 'square_neumann', n = 6, dim = 2 )
       u_gamma = sempy.BoundaryFunction( X )
       u_gamma.set_value( value = 3.0, physical_boundary = 2 )
      
    '''
    def __init__(self, Space,mask='yes'):
        self.Space = Space
        self.mask = mask
        if self.Space.dim == 1:
            self.basis_coeff =  np.zeros((self.Space.noe_bc),float)
            self.basis_coeff[:] = 0.0
            
        if self.Space.dim == 2:
            self.basis_coeff =  np.zeros((self.Space.noe_bc, \
                                          self.Space.n),float)
            self.basis_coeff[:,:] = 0.0
        
        if self.Space.dim == 3:
            self.basis_coeff =  np.zeros((self.Space.noe_bc, \
                                          self.Space.n,self.Space.n),float)
            self.basis_coeff[:,:,:] = 0.0
            
    def set_value(self,value,physical_boundary):
        '''
        Asigns *value* to the basis coefficients of the :class:`BoundaryFunction`. 
        
        :param value: Value to be assigned to the given boundary.
        :type value: Numpy array or floating point
        :param physical_boundary: Index of the physical boundary.
        :type physical_boundary: Integer
        '''
        
        if self.Space.dim == 1:
            val = np.zeros((self.Space.noe, self.Space.n),float)
            val[:,:] = value
            val_global = self.Space.local_to_global(val)
            #print 'obbs: check set_value in class BoundaryFunction'
            if self.mask == 'yes':
                val_global = self.Space.mask(val_global)
    
            for k in range(self.Space.noe_bc):
                [i, phys, bc_type, element_edge] = self.Space.theta_phys_bc(k)
                if phys == physical_boundary:
                    self.basis_coeff[k] = val_global[i]
                            
        if self.Space.dim == 2:
            val = np.zeros((self.Space.noe, self.Space.n, \
                            self.Space.n),float)
            val[:,:,:] = value
            val_global = self.Space.local_to_global(val)
            if self.mask == 'yes':
                val_global = self.Space.mask(val_global)
        
            for k in range(self.Space.noe_bc):
                for m in range(self.Space.n):
                    [i, phys, bc_type, element_edge] = self.Space.theta_phys_bc(k,m)
                    #print 'phys=',phys
                    #print 'physical_boundary=',physical_boundary
                    if phys == np.int(physical_boundary):
                        self.basis_coeff[k,m] = val_global[i]
                        
        if self.Space.dim == 3:
            val = np.zeros((self.Space.noe, self.Space.n, \
                            self.Space.n, self.Space.n),float)
            val[:,:,:,:] = value
            val_global = self.Space.local_to_global(val)
            if self.mask == 'yes':
                val_global = self.Space.mask(val_global)
        
            for k in range(self.Space.noe_bc):
                for m in range(self.Space.n):
                    for n in range(self.Space.n):
                        [i, phys, bc_type, element_edge] = self.Space.theta_phys_bc(k,m,n)
                        if phys == physical_boundary:
                            self.basis_coeff[k,m,n] = val_global[i]
        
    
        
class Function:
    '''
    An instance of this class represents a function in the discrete space.  
    
    .. .. math::
       X_N=\{v\in H^1(\Omega)\ |\ v|_{\Omega^k}\in\mathbb P_N(\Omega^k),\ 
             k=0,\ldots, \\text{NOE}-1 \}
             
       Y_N =\{ q\in L^2(\Omega)\ |\ q|_{\Omega^k}\in\mathbb P_N(\Omega^k),\ 
             k=0,\ldots, \\text{NOE}-1 \}
             
          
    :param Space: The function space.
    :type Space: :class:`sempy.Space`, :class:`sempy.precond.SpaceFEM` or
                 :class:`sempy.fluid.SpaceGL` 
    :param array basis_coeff: Basis coefficients of the function in local data 
                        representation.  
    :param string filename: The name of a file to dump data. 
    
    A function :math:`v` is defined as:
            
    .. math::
    
       \\forall v\in X_N,\quad v(\mathbf x)= 
       \sum_{i=0}^{\\text{dof}-1}\phi_i(\mathbf x)v_i
       
    where :math:`v_i` are the basis coefficients in global data 
    representation and the basis functions 
    :math:`\underline\phi\in\mathbb R^{\\text{dof}}` 
    have the property:
    
    .. math::
       \phi_j(\mathbf x_i)=\delta_{ij}
       
    In a local data structure the function is represented as:
    
    .. math::
       1D:&\quad v(\\xi)|_{\Omega^k}=\sum_{m=0}^N l_m(\\xi)v_{km}
       
       2D:&\quad v(\\xi,\\eta)|_{\Omega^k}=\sum_{m=0}^N \sum_{n=0}^N
                                          l_m(\\xi) l_n(\\eta) v_{kmn}
       
       3D:&\quad v(\\xi,\\eta,\\zeta)|_{\Omega^k}=\sum_{m=0}^N \sum_{n=0}^N\sum_{o=0}^N
                                          l_m(\\xi) l_n(\\eta) l_o(\\zeta)v_{kmno}
    
    **Example**::
    
      import sempy
      
      X = sempy.Space( 'square',n = 3, dim = 2 )
      u = sempy.Function( X, basis_coeff = X.x * X.x )
      
    '''
    def __init__( self, Space, basis_coeff = 0.0, 
                  filename = None, time = 0.0 ):
        
        self.Space = Space
        self.filename = filename
        self.file_count = 0
        self.time = time
               
        if self.Space.dim == 1:
            self.basis_coeff =  np.zeros((self.Space.noe, self.Space.n),
                                          float)
            self.basis_coeff[:,:] = basis_coeff
                        
        if self.Space.dim == 2:
            self.basis_coeff =  np.zeros((self.Space.noe, self.Space.n, 
                                          self.Space.n),float)
            self.basis_coeff[:,:,:] = basis_coeff
            
        if self.Space.dim == 3:
            self.basis_coeff =  np.zeros((self.Space.noe, self.Space.n, 
                                          self.Space.n, self.Space.n),float)
            self.basis_coeff[:,:,:,:] = basis_coeff
           
    #def at_boundary(self):
    #    '''
    #    Value at the boundary.
    #    '''
    #    return self.Space.boundary_value(self.coeff)
    def increment(self,h):
        """
        Increments the :attr:`sempy.Function.time` 
        attribute with :literal:`h`.

        >>> u.increment(0.1)

        """
        self.time += h
            
    def set_bc(self,value,physical_boundary):
        '''
        Sets the values of the Dirichlet boundary condition.
        
        :param value: The value to be assigned to the boundary. If 
                      :literal:`value` is given as an array it has to be 
                      given in local data representation. 
        :type value: array or floating point
        :param physical_boundary: The part of the boundary to recieve the value. 
        :type physical_boundary: integer
        
        Usage::
        
          import sempy
      
          X = sempy.Space( 'square', n = 3, dim = 2)
          u = sempy.Function( X, basis_coeff = X.x * X.x )
    
          u.set_bc( 1.0, 1 )
          u.set_bc( 1.0, 2 )
          u.plot_wire()
        '''
        if self.Space.type == 'GL':
            print 'Warning: cannot assign strong BC to a function in SpaceGL'
            return
        
        if self.Space.dim == 1:
            val = np.zeros((self.Space.noe, self.Space.n),float)
            val[:,:] = value
            val_global = self.Space.local_to_global(val)
            coeff_global = self.Space.local_to_global(self.basis_coeff)
        
            for k in range(self.Space.noe_bc):
                [i, phys, bc_type, element_edge] = self.Space.theta_phys_bc(k)
                if phys == physical_boundary:
                    coeff_global[i] = val_global[i]
                        
        if self.Space.dim == 2:
            val = np.zeros((self.Space.noe, self.Space.n, self.Space.n),float)
            val[:,:,:] = value
            val_global = self.Space.local_to_global(val)
            coeff_global = self.Space.local_to_global(self.basis_coeff)
        
            for k in range(self.Space.noe_bc):
                for m in range(self.Space.n):
                    [i, phys, bc_type, element_edge] = self.Space.theta_phys_bc(k,m)
                    if phys == physical_boundary:
                        coeff_global[i] = val_global[i]
        
        if self.Space.dim == 3:
            val = np.zeros((self.Space.noe, self.Space.n, self.Space.n, self.Space.n),float)
            val[:,:,:,:] = value
            val_global = self.Space.local_to_global(val)
            coeff_global = self.Space.local_to_global(self.basis_coeff)
        
            for k in range(self.Space.noe_bc):
                for m in range(self.Space.n):
                    for n in range(self.Space.n):
                        [i, phys, bc_type, element_edge] = self.Space.theta_phys_bc(k,m,n)
                        if phys == physical_boundary:
                            coeff_global[i] = val_global[i]
                        
        self.basis_coeff = self.Space.mapping_q(coeff_global)
    
    def set_value(self,value,physical_domain='none'):
        '''
        Assigns values to the basis coefficients. 
        
        :param value: Value to be assigned to the basis coefficients. 
        :type value: array or a floating number
        :param physical_domain: The index number of the physical domain, as 
                                given in the Gmsh geo file, 
                                to recieve the value. 
                                The default value is *none* in which the 
                                whole domain recieves the value. 
        :type physical_domain: integer
        
        Usage::
        
           import sempy
      
           X = sempy.Space( 'square', n = 3, dim = 2 )
           u = sempy.Function( X )
    
           u.set_value( 4.0 )
           u.plot_wire()
        
        '''
        if self.Space.dim == 1:
            val = np.zeros((self.Space.noe, self.Space.n),float)
            val[:,:] = value
            if physical_domain == 'none':
                self.basis_coeff[:,:]=val[:,:]
            else:
                for k in range(self.Space.noe):
                    i,phys=self.Space.theta_phys(k,0)
                    if phys == physical_domain:
                        self.basis_coeff[k,:]=val[k,:]
                        
        if self.Space.dim == 2:
            val = np.zeros((self.Space.noe, self.Space.n, self.Space.n),float)
            val[:,:,:] = value
            if physical_domain == 'none':
                self.basis_coeff[:,:,:]=val[:,:,:]
            else:
                for k in range(self.Space.noe):
                    i,phys=self.Space.theta_phys(k,0,0)
                    if phys == physical_domain:
                        self.basis_coeff[k,:,:]=val[k,:,:]
        
        if self.Space.dim == 3:
            val = np.zeros((self.Space.noe, self.Space.n, self.Space.n, self.Space.n),float)
            val[:,:,:,:] = value
            if physical_domain == 'none':
                self.basis_coeff[:,:,:,:]=val[:,:,:,:]
            else:
                for k in range(self.Space.noe):
                    i,phys=self.Space.theta_phys(k,0,0,0)
                    if phys == physical_domain:
                        self.basis_coeff[k,:,:,:]=val[k,:,:,:]
                    
        
    def glob(self):
        '''
        A method for grabbing the basis coefficients in global data structure. 
        
        :returns: The basis coefficients in global data structure. 
        :rtype: Numpy array
        
        Usage::
        
          import sempy
      
          X = sempy.Space( 'square',n = 5, dim = 2 )
          u = sempy.Function( X )
    
          u_glob = u.glob() 
        '''
        #if self.space == 'L2':
        #    print '---> error in TrialFunction.glob()'
        #    print '---> Cannot convert a function in L2 to'
        #    print '---> global data structure.'
        #else:
        return self.Space.local_to_global(self.basis_coeff)
    
    def quadrature(self):
        '''
        Computes the value :math:`\\kappa` of the integral 
        
        .. math::
           \\kappa = \\int_\\Omega f\\,d\\Omega 
           
        for :math:`f\\in L^2(\\Omega)` with GLL or GL quadrature rules, 
        depending of the corresponding space. 
        
        :returns: Value :math:`\\kappa` of the integral.
        
        Usage::
           
             import sempy
             
             X = sempy.Space( filename = 'square', n = 5, dim = 2 )
    
             u = sempy.Function( X, basis_coeff = X.x ) 
             print 'kappa=', u.quadrature()
        '''
        return self.Space.quadrature(self.basis_coeff)
        
    def l2_norm(self):
        """Computes the :math:`L^2` - norm
           
           .. math::
              ||u||_{L^2(\\Omega)}=\\bigg(\\int_{\\Omega} u^2d\\Omega\\bigg)^{1/2}
              
           using GLL or GL quadrature depending on the space. 
           
           :returns: The :math:`L^2` - norm as a floating number.
           
           Usage::
              
              import sempy
              
              X = sempy.Space( filename = 'square', n = 11, dim = 2 )
              
              u = sempy.Function( X, basis_coeff = X.x )
              print 'L2 norm=', u.l2_norm()

        """
        return self.Space.l2_norm( self.basis_coeff )
        
    def h1_norm(self):
        '''
        Computes the :math:`H^1(\\Omega)` - norm
           
        .. math::
           ||u||_{H^1(\\Omega)}=\\bigg(\\int_{\\Omega}\\bigg( u^2 + 
           \\sum_{i=1}^d\\bigg( \\frac{\\partial u}{\\partial x_i}\\bigg)^2
           \\bigg)
           d\\Omega\\bigg)^{1/2}
              
        using GLL quadrature. 
           
        :returns: The :math:`H^1` - norm as a floating number.
           
        Usage::
              
              import sempy
              
              X = sempy.Space( filename = 'square', n = 11, dim = 2 )
              
              u = sempy.Function( X, basis_coeff = X.x )
              print 'H1 norm=', u.h1_norm()
        
        '''
        if self.Space.dim == 1:
            u_x = self.grad()
            s_1 = self.Space.quadrature( self.basis_coeff * self.basis_coeff )
            s_2 = self.Space.quadrature( u_x * u_x )
            alpha = np.sqrt( s_1 + s_2 )
            return alpha
        
        if self.Space.dim == 2:
            u_x, u_y = self.grad()
            s_1 = self.Space.quadrature( self.basis_coeff * self.basis_coeff )
            s_2 = self.Space.quadrature( u_x * u_x )
            s_3 = self.Space.quadrature( u_y * u_y )
            alpha = np.sqrt( s_1 + s_2 + s_3 )
            return alpha
        
        if self.Space.dim == 3:
            u_x, u_y, u_z = self.grad()
            s_1 = self.Space.quadrature( self.basis_coeff * self.basis_coeff )
            s_2 = self.Space.quadrature( u_x * u_x )
            s_3 = self.Space.quadrature( u_y * u_y )
            s_4 = self.Space.quadrature( u_z * u_z )
            alpha = np.sqrt( s_1 + s_2 + s_3 + s_4 )
            return alpha
            
    def h1_semi_norm(self):
        '''
        Computes the :math:`H^1(\\Omega)` - semi-norm
           
        .. math::
           ||u||_{H^1(\\Omega)}=\\bigg(\\int_{\\Omega}
           \\sum_{i=1}^d\\bigg( \\frac{\\partial u}{\\partial x_i}\\bigg)^2
           d\\Omega\\bigg)^{1/2}
              
        using GLL quadrature. 
           
        :returns: The :math:`H^1` - semi-norm as a floating number.
           
        Usage::
              
              import sempy
              
              X = sempy.Space( filename = 'square', n = 11, dim = 2 )
              
              u = sempy.Function( X, basis_coeff = X.x )
              print 'H1 semi-norm=', u.h1_semi_norm()
        
        '''
        if self.Space.dim == 1:
            u_x = self.grad()
            s_1 = self.Space.quadrature( u_x * u_x )
            alpha = np.sqrt( s_1 )
            return alpha
        
        if self.Space.dim == 2:
            u_x, u_y = self.grad()
            s_1 = self.Space.quadrature( u_x * u_x )
            s_2 = self.Space.quadrature( u_y * u_y )
            alpha = np.sqrt( s_1 + s_2 )
            return alpha
        
        if self.Space.dim == 3:
            u_x, u_y, u_z = self.grad()
            s_1 = self.Space.quadrature( u_x * u_x )
            s_2 = self.Space.quadrature( u_y * u_y )
            s_3 = self.Space.quadrature( u_z * u_z )
            alpha = np.sqrt( s_1 + s_2 + s_3 )
            return alpha
            
        
    def grad(self):
        '''
        Computes the gradient vector of :math:`u`, i.e
        
        
        .. math::
        
           \\text{1D}:\quad \\nabla u= \partial u/\partial x
           
           \\text{2D}:\quad \\nabla u= (\partial u/\partial x,\partial u/\partial y)^T
           
           \\text{3D}:\quad \\nabla u= (\partial u/\partial x,\partial u/\partial y,\partial u/\partial z)^T
           
        :returns: The components of the gradient vector. 
        
        Usage::
          
          import sempy
          
          X = sempy.Space( 'square', n = 3, dim = 2 )
          u = sempy.Function( X, basis_coeff = X.x * X.x )
    
          # Calculate the gradient vector
          u_x, u_y = u.grad()
          u.set_value( u_x )
          u.plot_wire()
        
        '''
        #if self.Space.dim == 1:
        #    if self.Space.type == 'GL':
        #        print 'Warning: The gradient method is not implemented'
        #        print 'for functions in SpaceGL.'
        #        return
        #    else:
        #        u_x = sempy.space.space_f90.gradient_vector_1d(
        #                                    self.basis_coeff,
        #                                    self.Space.jac,
        #                                    self.Space.D)
        #        return u_x
        # 
        #if self.Space.dim == 2:
        #    if self.Space.type == 'GL':
        #        print 'Warning: The gradient method is not implemented'
        #        print 'for functions in SpaceGL.'
        #        return
        #    else:
        #        u_x, u_y = sempy.space.space_f90.gradient_vector_2d(
        #                                            self.basis_coeff,
        #                                            self.Space.x,
        #                                            self.Space.y,
        #                                            self.Space.jac,
        #                                            self.Space.D)
        #        return u_x, u_y
        # 
        #if self.Space.dim == 3:
        #    if self.Space.type == 'GL':
        #        print 'Warning: The gradient method is not implemented'
        #        print 'for functions in SpaceGL.'
        #        return
        #    else:
        #        #u_x, u_y, u_z = sempy.space.space_f90.gradient_vector_3d(
        #        #                self.basis_coeff,
        #        #                self.Space.jac,self.Space.D,
        #        #                self.Space.g11,self.Space.g12,
        #        #                self.Space.g13,self.Space.g21,
        #        #                self.Space.g22,self.Space.g23,
        #        #                self.Space.g31,self.Space.g32,
        #        #                self.Space.g33 )
        #        #print u_x
        return self.Space.gradient_vector( self.basis_coeff )
        
        #if self.Space.type == 'GL':
        #    print 'Warning: The gradient method is not implemented'
        #    print 'for functions in SpaceGL.'
        #    return
        #else:
        #    return self.Space.gradient_vector( self.basis_coeff )
        
    def hyper_filter(self,alpha=0.25):
        '''
        A filter for hyperbolic problems, see e.g. Deville et al. 
        
        .. math::
           u = F_\\alpha \\tilde u
           
        where the filter is
        
        .. math::
           F_\\alpha := \\alpha \\Pi_{N-1} + (1-\\alpha) I_N^N,\quad 
           \\Pi_{N-1} = I_{N-1}^N I_{N}^{N-1}
           
        Here, the weighting factor :math:`\\alpha` is typically in the 
        range :math:`0.05<\\alpha<0.3`. 
         
        :param alpha: Weighting factor.
        
        Usage::
        
          import sempy
          
          X = sempy.Space( 'square', n = 7, dim = 2 )
          u = sempy.Function( X, basis_coeff = X.x * X.x )
          
          u.hyper_filter( alpha = 0.3 )
          
        '''
        if self.Space.type == 'GL':
            print 'Warning: hyper_filter not applicable to functions in SpaceGL.'
            return
        if self.Space.n < 3:
            #return self.basis_coeff
            return
        else:
            #return self.Space.hyper_filter(self.basis_coeff,alpha)
            self.basis_coeff = self.Space.hyper_filter( self.basis_coeff,
                                                        alpha )
            return
    def plot(self,legend=True,filled=True,resolution=20.0,
             show_elements=True,font_size=20, show = True,
             show_boundary=False,
             lower='none',upper='none' ):
        if self.Space.dim==2:
            self.__plot_2d(legend=legend,filled=filled,
                           resolution=resolution,
                           show_elements=show_elements,
                           font_size=font_size,
                           show = show,
                           show_boundary=show_boundary,
                           lower=lower,upper=upper )
                
    def __plot_2d(self,legend=True,filled=True,resolution=20.0,
             show_elements=True,font_size=20, show = True,
             show_boundary=False,
             lower='none',upper='none' ):
        '''
        Plots the trial function. 
        
        Usage::
        
          u = Function(Space)
          u.plot()
          
        '''
        #rcParams['legend.fontsize']=3
        #fig=plt.figure()
        #ax=axes3d.Axes3D(fig)
        #for k in range(self.Space.noe):
        #    ax.plot_wireframe(self.Space.x[k,:,:],self.Space.y[k,:,:],self.coeff[k,:,:])
        #    #ax.plot_surface(self.x[k,:,:],self.y[k,:,:],u[k,:,:],rstride=1, cstride=1,cmap=cm.jet)
        #ax.set_xlabel(r'$x$')
        #ax.set_ylabel(r'$y$')
        #ax.set_zlabel(r'$u$')
        #plt.show()
        
        if self.Space.type == 'GL':
            u_i = self.Space.interpolation_gll(self.basis_coeff)
            x_i = self.Space.interpolation_gll(self.Space.x)
            y_i = self.Space.interpolation_gll(self.Space.y)
        else:
            n_int = 17
            u_i = self.Space.interpolation(self.basis_coeff,  n_int)
            x_i = self.Space.interpolation(self.Space.x, n_int)
            y_i = self.Space.interpolation(self.Space.y, n_int)
        
        extent_2 = (x_i.min(),x_i.max(),y_i.min(),y_i.max())
        
        fig_number = np.random.randint(0,101)
        
        #a_x=35.0/5.0
        #b_y=20.0/5.0
        #a_x=14.0/1.5#35.0/5.0
        #b_y=7.0/1.5#20.0/5.0
        #fig = plt.figure(fig_number,figsize=[ a_x, b_y ])
        fig = plt.figure(fig_number)  
    
        ax = fig.add_subplot(111)

        #levels=np.arange(-0.2,1.01,0.02)
        #beta = (abs(self.coeff.min())+abs(self.coeff.max()))/20.0
        #low  = self.coeff.min()-abs(self.coeff.min())*0.15
        #high = self.coeff.max()+self.coeff.max()*0.15
        #beta = (abs(low)+abs(high))/40.0
        if lower == 'none':
            low = u_i.min()-abs(u_i.min())#25
        else:
            low = lower
        #print 'low=',low
        if upper == 'none':
            high = u_i.max()#+abs(u_i.max())*0.12#25
        else:
            high = upper 
        beta   = (abs(low)+abs(high))/float(resolution)#40.0
        levels = np.arange(low-beta,high+beta,beta)
        #print levels
                
        if len(levels) == 0:
            levels= np.arange(-0.1,0.11,.01)
        #levels_new=np.zeros((len(levels)+1), float)
        #levels_new[0]=-0.13625
        #for i in range(len(levels)):
        #    levels_new[i+1]=levels[i]
        #del levels
        #levels=levels_new
        # Contour plot
        for k in range(self.Space.noe):
            if filled:
                plt.contourf(x_i[k,:,:],y_i[k,:,:],u_i[k,:,:],levels,\
                         cmap=cm.get_cmap('jet',len(levels)-1))#,extent=extent_2)
            else:
                plt.contour(x_i[k,:,:],y_i[k,:,:],u_i[k,:,:],levels,
                            #cmap=cm.get_cmap('jet',len(levels)-1),
                            linewidths=1,#)#,colors='b',linestyles='solid')
                            colors='k',linestyles='solid')
        #Elements
        thickness = 0.5
        if show_elements:
            if self.Space.type=='GL':
                n = self.Space.Space.n
                for k in range(self.Space.noe):
                    #for m in range(n):
                    plt.plot( self.Space.Space.x[k,:,0],
                              self.Space.Space.y[k,:,0],
                              color = 'k', linewidth = thickness)
                    plt.plot( self.Space.Space.x[k,:,n-1],
                              self.Space.Space.y[k,:,n-1],
                              color = 'k', linewidth = thickness)
                    plt.plot( self.Space.Space.x[k,0,:],
                              self.Space.Space.y[k,0,:],
                              color = 'k', linewidth = thickness)
                    plt.plot( self.Space.Space.x[k,n-1,:],
                              self.Space.Space.y[k,n-1,:],
                              color = 'k', linewidth = thickness)
            else:
                n = self.Space.n
                for k in range(self.Space.noe):
                    #for m in range(n):
                    plt.plot(self.Space.x[k,:,0],self.Space.y[k,:,0],
                         color='k',linewidth=thickness)
                    plt.plot(self.Space.x[k,:,n-1],self.Space.y[k,:,n-1],
                         color='k',linewidth=thickness)
                    plt.plot(self.Space.x[k,0,:],self.Space.y[k,0,:],
                         color='k',linewidth=thickness)
                    plt.plot(self.Space.x[k,n-1,:],self.Space.y[k,n-1,:],
                         color='k',linewidth=thickness)
        if legend:
            #rcParams['legend.fontsize']= font_size
            cbar = plt.colorbar( shrink=0.6)#,
                                 #orientation='horizontal')
            #cb = colorbar() # grab the Colorbar instance
            for t in cbar.ax.get_yticklabels():
                t.set_fontsize(20)

        
        if show_boundary:
            X = self.Space
            xb = X.boundary_value( X.x )
            yb = X.boundary_value( X.y )
            for k in range(X.noe_bc):
                plt.plot(xb[k,:],yb[k,:],color='k')
            #for k in range(self.Space.noe_bc):
            #    plt.plot
        #ax.xaxis.set_major_locator(ticker.FixedLocator(
        #                     (0.2,0.4,0.6,0.8)))
        #ax.yaxis.set_major_locator(ticker.FixedLocator(
        #                     (0.2,0.4,0.6,0.8)))

        plt.xlabel(r'$x$', fontsize = font_size )
        plt.ylabel(r'$y$', fontsize = font_size )
        
        plt.axis([self.Space.x.min(),self.Space.x.max(),
                  self.Space.y.min(),self.Space.y.max()])
        plt.axes().set_aspect('equal')
        
        ax = plt.gca()
        for label in ax.xaxis.get_ticklabels():
            label.set_fontsize( font_size )
        for label in ax.yaxis.get_ticklabels():
            label.set_fontsize( font_size )
        
        if show:
            plt.show()
        
        
    def plot_wire( self, levels = 7, show = True ):
        '''
        Visualization of the function.
        
        :param levels: Number of levels in the contour plot (for 3D functions).
        :type levels: int
        '''
        # filename = 'none'
        if self.Space.dim == 1:
            fig = plt.figure()
            for k in range(self.Space.noe):
                plt.plot(self.Space.x[k,:],self.basis_coeff[k,:])
            plt.xlabel(r'$x$')
            plt.ylabel(r'$u$')
            #if filename != 'none':
            #    print '---> filename=',filename
            #    plt.savefig(filename)
            if show:
                plt.show()
                
        if self.Space.dim == 2:
            fig = plt.figure()
            ax = axes3d.Axes3D(fig)
            for k in range(self.Space.noe):
                ax.plot_wireframe( self.Space.x[k,:,:], self.Space.y[k,:,:],
                                   self.basis_coeff[k,:,:])
            ax.set_xlabel(r'$x$')
            ax.set_ylabel(r'$y$')
            ax.set_zlabel(r'$u$')
            #if filename != 'none':
            #    print '---> filename=',filename
            #    plt.savefig(filename)
            plt.show()
            
        if  self.Space.dim == 3:
            self.__plot_wire_3d(levels)#contours=[0.01,0.02]):
    
    def __plot_wire_3d( self, levels ):#contours=[0.2,0.4,0.6,0.8]):
        try:
            import enthought.mayavi
        except:
            print 'No mayavi python module installed.'
            print '3D plotting not possible.'
            return
        #X_fem = sempy.precond.SpaceFEM(self.Space)
        #if self.SpaceFEM == 'none':
        #    print 'Error: Supply a SpaceFEM instance to facilitate 3D plotting'
        #    return
        
        u_max = self.basis_coeff.max()
        u_min = self.basis_coeff.min()
        du = ( u_max - u_min ) / float( levels )
        cont =np.arange( u_min, u_max, du )
        contours=['asdf']
        for i in range(len(cont)):
            contours.append(cont[i])
        contours.remove('asdf') 

        v_fem = sempy.Function(self.Space,filename='temp_file')
        v_fem.basis_coeff = self.basis_coeff
        #v_fem.basis_coeff = self.SpaceFEM.sem_to_fem(u.basis_coeff)
        v_fem.to_file()

        vtkFile='temp_file0.vtk'

        # Create the MayaVi engine and start it.
        from enthought.mayavi.core.api import Engine
        engine = Engine()
        engine.start()
        scene = engine.new_scene()

        # Read in VTK file and add as source
        from enthought.mayavi.sources.vtk_file_reader import VTKFileReader
        reader = VTKFileReader()
        reader.initialize(vtkFile)
        engine.add_source(reader)

        ## Add Surface Module
        #from enthought.mayavi.modules.surface import Surface
        #surface = Surface()
        #engine.add_module(surface)

        # Axes
        from enthought.mayavi.modules.axes import Axes
        axes = Axes()
        engine.add_module(axes)

        # Outline
        from enthought.mayavi.modules.outline import Outline
        outline = Outline()
        engine.add_module(outline)

        # IsoSurface
        from enthought.mayavi.modules.api import IsoSurface
        iso = IsoSurface()
        engine.add_module(iso)
        #iso.contour.contours = [-0.4, 0.0, 0.4]
        #iso.contour.contours = [-0.9, 0.45, 0.0, 0.45, 0.9]
        iso.contour.contours = contours#[0.01, 0.02, 0.03]
         #iso.contour.filled.contours = True
        # Make them translucent.
        #iso.actor.property.opacity = 0.9
        # Show the scalar bar (legend).
        iso.module_manager.scalar_lut_manager.show_scalar_bar = True

        # Move the camera
        scene.scene.camera.elevation(-45)

        # Save scene to image file
        #scene.scene.save_png('figures/poisson_3d_plot.png')

        # Create a GUI instance and start the event loop.
        # This stops the window from closing
        from enthought.pyface.api import GUI
        gui = GUI()
        gui.start_event_loop()
        
            
    def to_file_old(self):
        # Filename
        #if self.file_count == 0:
        #    _filename = self.filename + '.vtk'
        #    self.file_count=self.file_count+1
        #else:
        _filename = self.filename + str(self.file_count) + '.vtk'
        self.file_count=self.file_count+1
        print '_filename=',_filename
        
        # open file
        f = open(_filename,'w')
        f.write('# vtk DataFile Version 3.0\n')
        f.write('Sempy data file\n')
        f.write('ASCII\n')
        f.write('DATASET UNSTRUCTURED_GRID\n')
        
        #--- Points
        f.write('\n')
        f.write('POINTS ')
        a = np.zeros(1,np.int)
        a[0] = self.Space.dof
        f.write(np.array_str(a[0]))
        f.write(' float\n')
        
        xx = self.Space.local_to_global(self.Space.x)
        yy = self.Space.local_to_global(self.Space.y)
        zz = self.Space.local_to_global(self.Space.z)

        for k in range(self.Space.dof):
            f.write(np.array_str(xx[k]))
            f.write(' ')
            f.write(np.array_str(yy[k]))
            f.write(' ')
            f.write(np.array_str(zz[k]))
            f.write(' \n')

        #--- Cells
        f.write('\n')
        f.write('CELLS ')
        a=np.zeros(1,np.int)
        a[0]=self.Space.noe
        f.write(np.array_str(a[0]))
        f.write(' ')
        a=np.zeros(1,np.int)
        a[0]=self.Space.noe*(self.Space.n*self.Space.n*self.Space.n + 1)
        f.write(np.array_str(a[0]))
        f.write('\n')
        for k in range(self.Space.noe):
            f.write('8 ')
            f.write(np.array_str(self.Space.theta[k,0,0,0]))
            f.write(' ')
            f.write(np.array_str(self.Space.theta[k,1,0,0]))
            f.write(' ')
            f.write(np.array_str(self.Space.theta[k,1,0,1]))
            f.write(' ')
            f.write(np.array_str(self.Space.theta[k,0,0,1]))
            f.write(' ')
            f.write(np.array_str(self.Space.theta[k,0,1,0]))
            f.write(' ')
            f.write(np.array_str(self.Space.theta[k,1,1,0]))
            f.write(' ')
            f.write(np.array_str(self.Space.theta[k,1,1,1]))
            f.write(' ')
            f.write(np.array_str(self.Space.theta[k,0,1,1]))
            f.write(' \n')
            
        #--- Cell types
        f.write('\n')
        f.write('CELL_TYPES ')
        a=np.zeros(1,np.int)
        a[0]=self.Space.noe
        f.write(np.array_str(a[0]))
        f.write('\n')
        a[0]=12
        for k in range(self.Space.noe):
            f.write(np.array_str(a[0]))
            f.write(' \n')
        f.write('\n')
        
        #--- Scalar
        #if self.file_count > 0:
        _u = self.Space.local_to_global(self.basis_coeff)
        
        f.write('POINT_DATA ')
        a=np.zeros(1,np.int)
        a[0]=self.Space.dof
        f.write(np.array_str(a[0]))
        f.write('\n')
        f.write('SCALARS scalars float\n')
        f.write('LOOKUP_TABLE default\n')
        #a=np.zeros(1,np.float)
        #a[0]=3.0
        for k in range(self.Space.dof):
            f.write(np.array_str(_u[k]))
            #f.write(np.array_str(a[0]))
            f.write(' \n')
    
        f.close()
    
    def to_file(self):
        '''
        This method prints the basis coefficients to a 
        file. For 2D and 3D functions, the coefficients is written 
        into a file of the vtk file format. 1D functions are written 
        into a trivial .dat file with two columns.
        '''
        if not self.filename == None:
            if self.Space.dim == 1:
                self.__to_file_1d()
            if self.Space.dim == 2:
                self.__to_file_2d()
            if self.Space.dim == 3:
                self.__to_file_3d()
            
    def __to_file_1d(self):
        '''
        Print the basis coefficients to a file in the vtk file format.
        '''
        # Filename
        #if self.file_count == 0:
        #    _filename = self.filename + '.vtk'
        #    self.file_count=self.file_count+1
        #else:
        
        _filename = self.filename + str(self.file_count) + '.dat'
        self.file_count=self.file_count+1
        print '_filename=',_filename
        
        # open file
        f = open(_filename,'w')
        
        xx = self.Space.local_to_global(self.Space.x)
        uu = self.Space.local_to_global(self.basis_coeff)
        
        # Write to file
        for k in range(self.Space.dof):
            f.write(np.array_str(xx[k]))
            f.write(' ')
            f.write(np.array_str(uu[k]))
            f.write(' \n')
            
        # Close file
        f.close()
        
    def __to_file_2d(self):
        '''
        Print the basis coefficients to a file in the vtk file format.
        '''
        # Filename
        #if self.file_count == 0:
        #    _filename = self.filename + '.vtk'
        #    self.file_count=self.file_count+1
        #else:
        
        _filename = self.filename + str(self.file_count) + '.vtk'
        self.file_count=self.file_count+1
        
        #print '_filename=',_filename
        
        # open file
        f = open(_filename,'w')
        f.write('# vtk DataFile Version 3.0\n')
        f.write('Sempy data file\n')
        f.write('ASCII\n')
        f.write('DATASET UNSTRUCTURED_GRID\n')
        
        #--- Points
        f.write('\n')
        f.write('POINTS ')
        a = np.zeros(1,np.int)
        a[0] = self.Space.dof
        f.write(np.array_str(a[0]))
        f.write(' float\n')
        
        
        xx = self.Space.fem.local_to_global(self.Space.fem.x)
        yy = self.Space.fem.local_to_global(self.Space.fem.y)
        zz = np.zeros((self.Space.dof),float)

        for k in range(self.Space.dof):
            f.write(np.array_str(xx[k]))
            f.write(' ')
            f.write(np.array_str(yy[k]))
            f.write(' ')
            f.write(np.array_str(zz[k]))
            f.write(' \n')

        #--- Cells
        f.write('\n')
        f.write('CELLS ')
        a = np.zeros(1,np.int)
        a[0] = self.Space.fem.noe
        f.write( np.array_str( a[0] ) )
        f.write( ' ' ) 
        a = np.zeros( 1, np.int )
        a[0] = self.Space.fem.noe*(self.Space.fem.n*self.Space.fem.n + 1)
        f.write( np.array_str( a[0] ) )
        f.write('\n')
        for k in range(self.Space.fem.noe):
            f.write('4 ')
            f.write(np.array_str(self.Space.fem.theta[k,0,0]))
            f.write(' ')
            f.write(np.array_str(self.Space.fem.theta[k,1,0]))
            f.write(' ')
            f.write(np.array_str(self.Space.fem.theta[k,0,1]))
            f.write(' ')
            f.write(np.array_str(self.Space.fem.theta[k,1,1]))
            f.write(' \n')
            
        #--- Cell types
        f.write( '\n' )
        f.write( 'CELL_TYPES ' )
        a = np.zeros( 1, np.int )
        a[0] = self.Space.fem.noe
        f.write( np.array_str( a[0] ) )
        f.write( '\n' )
        a[0] = 8
        for k in range(self.Space.fem.noe):
            f.write(np.array_str(a[0]))
            f.write(' \n')
        f.write('\n')
        
        #--- Scalar
        #if self.file_count > 0:
        __u = self.Space.fem.sem_to_fem(self.basis_coeff)
    
        _u = self.Space.fem.local_to_global(__u)#self.basis_coeff)
        
        f.write('POINT_DATA ')
        a = np.zeros(1,np.int)
        a[0] = self.Space.dof
        f.write( np.array_str(a[0]) )
        f.write('\n')
        f.write('SCALARS scalars float\n')
        f.write('LOOKUP_TABLE default\n')
        #a=np.zeros(1,np.float)
        #a[0]=3.0
        for k in range(self.Space.dof):
            f.write(np.array_str(_u[k]))
            #f.write(np.array_str(a[0]))
            f.write(' \n')
    
        f.close()
                
    def __to_file_3d(self):
        '''
        Print the basis coefficients to a file in the vtk file format.
        '''
        # Filename
        #if self.file_count == 0:
        #    _filename = self.filename + '.vtk'
        #    self.file_count=self.file_count+1
        #else:
        
        _filename = self.filename + str(self.file_count) + '.vtk'
        self.file_count=self.file_count+1
        #print '_filename=',_filename
        
        # open file
        f = open(_filename,'w')
        f.write('# vtk DataFile Version 3.0\n')
        f.write('Sempy data file\n')
        f.write('ASCII\n')
        f.write('DATASET UNSTRUCTURED_GRID\n')
        
        #--- Points
        f.write('\n')
        f.write('POINTS ')
        a = np.zeros(1,np.int)
        a[0] = self.Space.dof
        f.write(np.array_str(a[0]))
        f.write(' float\n')
        
        
        xx = self.Space.fem.local_to_global(self.Space.fem.x)
        yy = self.Space.fem.local_to_global(self.Space.fem.y)
        zz = self.Space.fem.local_to_global(self.Space.fem.z)

        for k in range(self.Space.dof):
            f.write(np.array_str(xx[k]))
            f.write(' ')
            f.write(np.array_str(yy[k]))
            f.write(' ')
            f.write(np.array_str(zz[k]))
            f.write(' \n')

        #--- Cells
        f.write('\n')
        f.write('CELLS ')
        a = np.zeros(1,np.int)
        a[0] = self.Space.fem.noe
        f.write( np.array_str( a[0] ) )
        f.write( ' ' ) 
        a = np.zeros( 1, np.int )
        a[0] = self.Space.fem.noe*(self.Space.fem.n*self.Space.fem.n*self.Space.fem.n + 1)
        f.write( np.array_str( a[0] ) )
        f.write('\n')
        for k in range(self.Space.fem.noe):
            f.write('8 ')
            f.write(np.array_str(self.Space.fem.theta[k,0,0,0]))
            f.write(' ')
            f.write(np.array_str(self.Space.fem.theta[k,1,0,0]))
            f.write(' ')
            f.write(np.array_str(self.Space.fem.theta[k,1,0,1]))
            f.write(' ')
            f.write(np.array_str(self.Space.fem.theta[k,0,0,1]))
            f.write(' ')
            f.write(np.array_str(self.Space.fem.theta[k,0,1,0]))
            f.write(' ')
            f.write(np.array_str(self.Space.fem.theta[k,1,1,0]))
            f.write(' ')
            f.write(np.array_str(self.Space.fem.theta[k,1,1,1]))
            f.write(' ')
            f.write(np.array_str(self.Space.fem.theta[k,0,1,1]))
            f.write(' \n')
            
        #--- Cell types
        f.write( '\n' )
        f.write( 'CELL_TYPES ' )
        a = np.zeros( 1, np.int )
        a[0] = self.Space.fem.noe
        f.write( np.array_str( a[0] ) )
        f.write( '\n' )
        a[0] = 12
        for k in range(self.Space.fem.noe):
            f.write(np.array_str(a[0]))
            f.write(' \n')
        f.write('\n')
        
        #--- Scalar
        #if self.file_count > 0:
        __u = self.Space.fem.sem_to_fem(self.basis_coeff)
    
        _u = self.Space.fem.local_to_global(__u)#self.basis_coeff)
        
        f.write('POINT_DATA ')
        a = np.zeros(1,np.int)
        a[0] = self.Space.dof
        f.write( np.array_str(a[0]) )
        f.write('\n')
        f.write('SCALARS scalars float\n')
        f.write('LOOKUP_TABLE default\n')
        #a=np.zeros(1,np.float)
        #a[0]=3.0
        for k in range(self.Space.dof):
            f.write(np.array_str(_u[k]))
            #f.write(np.array_str(a[0]))
            f.write(' \n')
    
        f.close()


if __name__ == "__main__":
    #import doctest
    #doctest.testmod()
    import sempy
      
    #X = sempy.Space('line',n = 6, dim = 1)
    #X = sempy.Space('square',n = 6, dim = 2)
    X = sempy.Space('cube',n = 4, dim = 3)
    




import sempy
from prk import PRK
import numpy as np


def bdf1():
    # First order.
    alpha =np.zeros((2),float)
    beta =np.zeros((2),float)
    alpha[1], alpha[0] = 1.0, -1.0
    beta[1] = 1.0
    levels=2
    return alpha,beta,levels

def bdf2():
    # second order
    alpha =np.zeros((3),float)
    beta =np.zeros((3),float)
    alpha[2], alpha[1], alpha[0] = 1.5, -2.0, 0.5
    beta[2] = 1.0
    levels=3
    return alpha,beta,levels

def bdf3():
    # Third order.
    alpha =np.zeros((4),float)
    beta =np.zeros((4),float)
    alpha[3],alpha[2],alpha[1],alpha[0]=11.0/6.0,-3.0,3.0/2.0,-1.0/3.0
    beta[3]=1.0
    levels=4
    return alpha,beta,levels

def bdf4():
    # Fourth order
    alpha =np.zeros((5),float)
    beta =np.zeros((5),float)
    alpha[4],alpha[3],alpha[2],alpha[1],alpha[0]=25.0/12.0,-4.0,3.0,-4.0/3.0,1.0/4.0
    beta[4] = 1.0
    levels=5
    return alpha,beta,levels

def __bc__(y,t):
    return y

class OIFS(PRK):
    '''
    Operator integration factor splitting (OIFS) method for the solution 
    of 
    
    .. math::
       M\\frac{dy}{dt}=Ly+N(y,t)
    
    where :math:`M` is the mass matrix, :math:`L` is a linear operator 
    and :math:`N(y,t)` is a non-linear function. 
    
    :param y: Initial condition
    :type y: A collection of :class:`sempy.Function`
    :param Time: An instance of the :class:`sempy.ode.Time` class.
    :kwargs: * **L** - Linear operator
             * **L_pre** - The FEM equivalent of L. Used to create 
               preconditioner. 
             * **force_function** - A function, possibly non-linear. 
               This is the function :math:`N(y,t)` in the equation above.
             * **bound_cond** - Strong BCs.
             * **order** (*int*) - Order of the method to be used. 
             * **explicit_steps** (*int*) - Number of steps with the explicit 
               solver. 
             * **init_steps** (*int*) - Number of inital steps to kickstart 
               the solution process. 
             * **linear_solver** - The linear solver. The default is 
               CG. If L is not SPD, then one can for example use bicgstab.
             * **iter_tol** (*float*) - Iterative tolerance.
    '''
    def __init__(self, y, Time, L = None, L_pre= None,
                 force_function = None, bound_cond = __bc__,
                 order = 1, explicit_steps = 1,
                 init_steps = 1, file_increment = 1000,
                 hyper_filter = False,call_back=None,
                 linear_solver = None, iter_tol = 1e-10,maxiter=1000):
        PRK.__init__( self, y, Time, L = L, L_pre= L_pre,
                      force_function = force_function, 
                      bound_cond = bound_cond, 
                      file_increment = file_increment,
                      hyper_filter = hyper_filter,call_back=call_back,
                      linear_solver = linear_solver, iter_tol = iter_tol,
                      maxiter=maxiter)
        
        self.explicit_steps = explicit_steps
        self.init_steps = init_steps
        self.order = order
        
        if self.order==1:
            [self.alpha,self.beta,self.levels]=bdf1()
        if self.order==2:
            [self.alpha,self.beta,self.levels]=bdf2()
        if self.order==3:
            [self.alpha,self.beta,self.levels]=bdf3()
        if self.order==4:
            [self.alpha,self.beta,self.levels]=bdf4()
            
    def solve(self):
        '''
        Solve the equations. 
        '''
        #-- Solution vectors
        y_vec = np.zeros( (self.vector_size), float )
        #
        y_tilde=[]
        for i in range(self.levels-1):
            y_temp=[]
            for j in range(self.number_of_eq):
                y_temp.append( sempy.Function(self.y0[j].Space,
                                              basis_coeff=0.0) )
            y_tilde.append( y_temp )
            del y_temp
        y=[]
        for i in range(self.levels):
            y_temp=[]
            for j in range(self.number_of_eq):
                y_temp.append( sempy.Function(self.y0[j].Space,
                                              basis_coeff=0.0) )
            y.append( y_temp )
            del y_temp
        # Init value
        for i in range(self.levels-1):
            for j in range(self.number_of_eq):
                y_tilde[i][j].basis_coeff=np.copy(self.y0[j].basis_coeff)
        for i in range(self.levels):
            for j in range(self.number_of_eq):
                y[i][j].basis_coeff=np.copy(self.y0[j].basis_coeff)
        #-- Explicit solver
        Time_local = sempy.ode.Time(self.Time.Space,
                        start_time = 0.0,
                        end_time = self.Time.h,
                        time_steps = self.explicit_steps)
        y_help = self.output_vector(self.y0,adopt_filename=False)
        erk = sempy.ode.PRK( y_help,Time_local,
                             force_function = self.force_function, 
                             bound_cond = self.bound_cond,
                             prk_method='erk4',
                             hyper_filter=self.hyper_filter,
                             file_increment = 100000)
        # -- Callback
        if not self.call_back == None:
            self.call_back(y[0],self.Time.time,0)
        #-- Initialize
        factor=0.0
        for i in range(0,self.levels-2):
            self.Time.increment()
            factor = factor + 1.0
            y_temp = self.__initialize__( y[i],
                                          (factor-1.0)*self.Time.h,
                                          factor*self.Time.h )
            for k in range(self.number_of_eq):
                y[i+1][k].basis_coeff = np.copy(y_temp[k].basis_coeff)
            del y_temp
            # -- Callback
            if not self.call_back == None:
                self.call_back(y[i+1],self.Time.time,i+1)
                
           
        #-- Time-stepping
        for i in range(self.levels-2,self.Time.time_steps):
            self.Time.increment()
            time_global = self.Time.time
            h = self.Time.h
            # Solve the non-linear problem
            for j in range(self.levels-1):
                Time_local = sempy.ode.Time(self.Time.Space,
                               start_time = time_global - \
                                     np.float(self.levels-j-1)*h,
                               end_time = time_global,
                               time_steps = \
                                   self.explicit_steps*(self.levels-j-1))
                # Initial condition
                for k in range(self.number_of_eq):
                    erk.y0[k].basis_coeff = np.copy(y[j][k].basis_coeff)
                erk.Time = Time_local
                # Solve the problem
                y_temp = erk.solve()
                
                for k in range(self.number_of_eq):
                    y_tilde[j][k].basis_coeff = np.copy(y_temp[k].basis_coeff)
                del y_temp 
            # Calculate right hand side for the linear problem
            b = self.__rhs__(y_tilde)
            # Helmholtz operator
            H = sempy.operators.MultipleOperators([self.M,self.L],
                      scaling_factor=[self.alpha[self.levels-1], 
                                    -self.Time.h*self.beta[self.levels-1]],
                      assemble='no').matrix
            # Solve linear system
            # Boundary conditions
            y[self.levels-1] = self.bound_cond(y[self.levels-1],
                                               self.Time.time)
            y_vec = self.component_to_vector( y[self.levels-1] )
            [y_vec,flag] = self.linear_solver.solve(H, b, y_vec)
            y[self.levels-1] = self.vector_to_component_x(y_vec,
                                                   y[self.levels-1])
            print 'iter=',flag
            # Update solution array
            for k in range(self.levels-1):
                for j in range(self.number_of_eq):
                    y[k][j].basis_coeff=np.copy(y[k+1][j].basis_coeff)
            # -- Callback
            if not self.call_back == None:
                self.call_back(y[self.levels-1],self.Time.time,i+1)
                
        return y[self.levels-1]
    
    def __initialize__(self,y_0,start,end):
        print 'initialize...'
        
        Time_local = sempy.ode.Time( self.Time.Space,
                             start_time = start,
                             end_time = end,
                             time_steps = self.init_steps )
        print 'local h=',Time_local.h
        y_t=[]
        for j in range(self.number_of_eq):
            y_t.append( sempy.Function(y_0[j].Space,
                                       basis_coeff=0.0) )
        for k in range(self.number_of_eq):
            y_t[k].basis_coeff = np.copy(y_0[k].basis_coeff)
        euler= OIFS( y_t, Time_local, 
                     L = self.L_in, L_pre= self.L_pre,
                     force_function = self.force_function, 
                     bound_cond = self.bound_cond,
                     order = 1, 
                     explicit_steps = self.explicit_steps,
                     linear_solver = self.linear_solver, 
                     iter_tol = self.iter_tol,
                     file_increment = 10000)
        y1 = euler.solve()
        del euler
        return y1
    
    def vector_to_component_x(self,y_vec,y_comp):
        #
        dof_1 = 0
        dof_2 = 0
        for i in range(self.number_of_eq):
            dof_1 = dof_2
            dof_2 = dof_2 + y_comp[i].Space.dof
            temp = y_comp[i].Space.mapping_q( y_vec[dof_1:dof_2] )
            y_comp[i].basis_coeff = np.copy(temp)
        return y_comp
            
    def __rhs__(self,y_tilde):
        F = np.zeros((self.vector_size),float)
        #s = np.zeros((self.vector_size),float)
        # Time derivative
        #for i in range(0,self.alpha.size-1):
        for i in range(self.levels-1):
            if abs(self.alpha[i]) > 0.0:
                y_vec = self.component_to_vector( y_tilde[i] )
                F = F - self.alpha[i] * ( self.M * y_vec )
        # weak BCs
        return F
        
        


'''
Sempy: A Python implementation of the spectral element method.
Copyright (C) 2011,  Stian Jensen.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
''' 

import numpy as np
import sempy
import sempy.space.space_f90 as space_f90
import sempy.operators.operators_f90 as operators_f90

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import axes3d
from matplotlib import cm

import time 

def MeshConvert3D(filename,elem_dof=3):
    '''
    Takes as input a gmsh file (filename) and gmsh element type (elem_dof).
    Returns: x, y, physical_names, theta, physical, theta_bc, physical_bc, dof, noe
    '''
    #print 'Reading gmsh file....'
    #--- Physical names
    f=open(filename,'r')
    i=0
    for line in f:
        i=i+1
        if line == "$PhysicalNames\n":
            name_start=i+1
        if line == "$EndPhysicalNames\n":
            name_end=i
    f.close()
    
    
    f=open(filename,'r')
    f_names=open('names.txt','w')
    i=0
    for line in f:
        i=i+1
        if i > name_start and i < name_end:
            f_names.write(line)
    f.close()
    f_names.close()
    f_names=open('names.txt','r')
    physical_names = np.genfromtxt(f_names,usecols=(0,1),dtype=int)
    f_names.close()
    #print 'physical_names='
    #print physical_names
    #print '**************'
    # Determining the bc type, i.e. "dir" (strong, Dirichlet) or "Nat" 
    # (natural like Robin and Neumann)
    f_names=open('names.txt','r')
    bc_type=['0']
    for line in f_names:
        #print line
        if int(line[0]) == 2:
            bc_type.append(line[5:8])
    f_names.close()
    bc_type.remove('0')
    #print 'bc_type=',bc_type
    #print 'physical_names=',physical_names
    #print '**************'
    
    #--- Locating the nodes and elements in the  gmsh-file:
    f=open(filename,'r')
    i=0
    for line in f:
        i=i+1
        if line == "$Nodes\n":
            nodes_start=i+1
        if line == "$EndNodes\n":
            nodes_end=i        
        if line == "$Elements\n":
            elements_start=i+1
        if line == "$EndElements\n":
            elements_end=i
    f.close()

    #--- Creating temporary files of nodes and elements
    f          = open(filename,'r')
    f_nodes    = open('nodes.txt','w')
    f_elements = open('elements.txt','w')

    i=0
    for line in f:
        i=i+1
        if i > nodes_start and i < nodes_end:
            f_nodes.write(line)
        if i > elements_start and i < elements_end:
            f_elements.write(line)
    f.close()
    f_nodes.close()
    f_elements.close()

    #--- Nodal points (x,y)
    f_nodes = open('nodes.txt','r')
    A = np.genfromtxt(f_nodes,dtype=float)
    f_nodes.close()
    x=A[:,1]
    y=A[:,2]
    z=A[:,3]
    #print 'A='
    #print A   
    #print 'x shape=',x.shape
    #--- Connectivity array (theta_bc)
    # Finding internal elements
    f_elements = open('elements.txt','r')
    A = np.genfromtxt(f_elements, dtype=int,
                      usecols=(0,1,2,3,4,5,6,7,8,9,10,11,12,13,14))
    f_elements.close()
    #print 'A.shape=',A.shape
    #print A
    # Removing the boundary elements
    #print 'elem_dof=',elem_dof
    if elem_dof == 2:
        element_type = 5 # gmsh type
    if elem_dof == 3:
        element_type = 12 # gmsh type
    for i in range(A.shape[0]):
        if A[i,1] == element_type:
            cutoff=i
            break
    #print 'cutoff=',cutoff
    
    # Boundary elements
    theta_bc=np.zeros((cutoff,elem_dof,elem_dof),int)
    physical_bc=np.zeros((cutoff),int)
    for i in range(cutoff):
        physical_bc[i]=A[i,3]
    #print 'physical_bc='
    #print physical_bc
    
    if elem_dof == 3:
        for i in range(cutoff):
            theta_bc[i,0,0] = A[i,6]-1
            theta_bc[i,1,0] = A[i,7]-1
            theta_bc[i,1,1] = A[i,8]-1
            theta_bc[i,0,1] = A[i,9]-1
            
    if elem_dof == 3:
        for i in range(cutoff):
            theta_bc[i,0,0] = A[i,6]-1
            theta_bc[i,1,0] = A[i,10]-1
            theta_bc[i,2,0] = A[i,7]-1
            theta_bc[i,2,1] = A[i,11]-1
            theta_bc[i,1,2] = A[i,12]-1
            theta_bc[i,0,1] = A[i,13]-1
            theta_bc[i,1,1] = A[i,14]-1
            theta_bc[i,2,2] = A[i,8]-1
            theta_bc[i,0,2] = A[i,9]-1
    #print 'theta_bc'
    #print theta_bc
    # Creating array of elements without boundary elements
    f_elements = open( 'elements.txt', 'r' )
    ver = np.fromstring( np.__version__, dtype = float, sep='.' )
    if ver[0] > 1.3:
        internal_elements = np.genfromtxt( f_elements, dtype = int, 
                                           skip_header = cutoff )
    else:
        internal_elements = np.genfromtxt( f_elements, dtype = int, 
                                           skiprows = cutoff )
    f_elements.close()

    #print 'internal_elements shape=',internal_elements.shape
    #print internal_elements

    k = internal_elements.shape[0]
    
    if elem_dof == 2:
        theta        = np.zeros((k,2,2,2),int)
        physical     = np.zeros((k),int)
        theta[:,0,0,0] = internal_elements[:,6]-1 # 0
        theta[:,1,0,0] = internal_elements[:,7]-1 # 1
        theta[:,1,1,0] = internal_elements[:,8]-1 # 2
        theta[:,0,1,0] = internal_elements[:,9]-1 # 3
        theta[:,0,0,1] = internal_elements[:,10]-1 # 4
        theta[:,1,0,1] = internal_elements[:,11]-1 # 5
        theta[:,1,1,1] = internal_elements[:,12]-1 # 6
        theta[:,0,1,1] = internal_elements[:,13]-1 # 7
        
        physical[:]  = internal_elements[:,3]
        
    if elem_dof == 3:
        theta        = np.zeros((k,3,3,3),int)
        physical     = np.zeros((k),int)
        theta[:,0,0,0] = internal_elements[:,6]-1 # 0
        theta[:,2,0,0] = internal_elements[:,7]-1 # 1
        theta[:,2,2,0] = internal_elements[:,8]-1 # 2
        theta[:,0,2,0] = internal_elements[:,9]-1 # 3
        
        theta[:,0,0,2] = internal_elements[:,10]-1 # 4
        theta[:,2,0,2] = internal_elements[:,11]-1 # 5
        theta[:,2,2,2] = internal_elements[:,12]-1 # 6
        theta[:,0,2,2] = internal_elements[:,13]-1 # 7
        
        theta[:,1,0,0] = internal_elements[:,14]-1 # 8
        theta[:,0,1,0] = internal_elements[:,15]-1 # 9
        theta[:,0,0,1] = internal_elements[:,16]-1 # 10
        theta[:,2,1,0] = internal_elements[:,17]-1 # 11
        
        theta[:,2,0,1] = internal_elements[:,18]-1 # 12
        theta[:,1,2,0] = internal_elements[:,19]-1 # 13
        theta[:,2,2,1] = internal_elements[:,20]-1 # 14
        theta[:,0,2,1] = internal_elements[:,21]-1 # 15
        
        theta[:,1,0,2] = internal_elements[:,22]-1 # 16
        theta[:,0,1,2] = internal_elements[:,23]-1 # 17
        theta[:,2,1,2] = internal_elements[:,24]-1 # 18
        theta[:,1,2,2] = internal_elements[:,25]-1 # 19
        
        theta[:,1,1,0] = internal_elements[:,26]-1 # 20
        theta[:,1,0,1] = internal_elements[:,27]-1 # 21
        theta[:,0,1,1] = internal_elements[:,28]-1 # 22
        theta[:,2,1,1] = internal_elements[:,29]-1 # 23
        
        theta[:,1,2,1] = internal_elements[:,30]-1 # 24
        theta[:,1,1,2] = internal_elements[:,31]-1 # 25
        theta[:,1,1,1] = internal_elements[:,32]-1 # 26
        
        physical[:]  = internal_elements[:,3]

    dof=x.shape[0]
    noe=theta.shape[0]
        
    return x, y, z, physical_names, theta, physical, \
           theta_bc, physical_bc, dof, noe, bc_type



class GmeshToSpectral3D():
    def __init__(self,filename,n=3):
        #print 'Creating Gmsh object ..'
        self.dim = 3
        self.n = n
        self.basic = sempy.Basic()
        self.D = self.basic.derivative_matrix_gll(self.n)
        x_glob, y_glob, z_glob, self.physical_names, self.theta, self.physical,\
        self.theta_bc, self.physical_bc, self.dof, self.noe, self.bc_type\
            = MeshConvert3D(filename,self.n)
        self.noe_bc      = self.theta_bc.shape[0]
        self.points, self.weights = \
                    self.basic.gauss_lobatto_legendre( self.n )
        self.x = self.mapping_q(x_glob)
        self.y = self.mapping_q(y_glob)
        self.z = self.mapping_q(z_glob)
        #print '... finished creating Gmsh object.'
        #self.jac = space_f90.geometric_3d(self.x, self.y, self.z, self.D)
        
        #print '**** noe =',self.noe
        #print '**** dof =',self.dof
        #print '**** n   =',self.n
        
    #def theta_phys_bc(self, k, m, n):
    #    i       = self.theta_bc[k,m,n]
    #    phys    = self.physical_bc[k]
    #    boundarycondtion_type = self.bc_type[phys-1]
    #    element_edge = 0#self.edge[k]
    #    return i, phys, boundarycondtion_type, element_edge
        
    #def theta_phys(self,k,m,n,o):
    #    i    = self.theta[k,m,n,o]
    #    phys = self.physical[k]
    #    return i, phys
        
    def mapping_q(self,v):
        # from global to local
        w = space_f90.mapping_q_3d(v,self.theta)
        #w=np.zeros((self.noe, self.n, self.n, self.n),float)
        #for k in range(self.noe):
        #    for m in range(self.n):
        #        for n in range(self.n):
        #            for o in range(self.n):
        #                i = self.theta[k,m,n,o]
        #                w[k,m,n,o]=v[i]
        return w
        
    #def mapping_qt(self,w):
    #    v = space_f90.mapping_qt_3d(w,self.theta,self.dof)
    #    return v
        
    #def local_to_global(self,w):
    #    v = space_f90.local_to_global_3d(w,self.theta,self.dof)
    #    return v
        
    #def quadrature(self,f):
    #    #kappa = space_f90.quadrature_3d(self.weights,self.jac,f)
    #    kappa=0.0
    #    for k in range(self.noe):
    #        for m in range(self.n):
    #            for n in range(self.n):
    #                for o in range(self.n):
    #                    kappa = kappa + self.weights[m]*self.weights[n]*self.weights[o]*\
    #                            self.jac[k,m,n,o]*f[k,m,n,o]
    #    return kappa
        
    #def l2_norm(self,u):
    #    norm = space_f90.l2norm_3d(self.weights,self.jac,u)
    #    return norm
    
    #def mask(self,w):
    #    '''
    #    This method masks nodal point values associated with a Dirichlet boundary condition.
    #    
    #    :param w: Numpy array in global data representation subject to masking.
    #    :returns: A masked numpy array in global data representation.
    #    
    #    '''
    #    for k in range(self.noe_bc):
    #        for m in range(self.n):
    #            for n in range(self.n):
    #                [i, name, bc_type, edge] = self.theta_phys_bc(k,m,n)
    #                if bc_type == 'Dir':
    #                    w[i] = 0.0
    #    return w


class SimpleMesh3D():
    def __init__(self):
        #print 'Creating Gmsh object ..'
        self.dim = 3
        self.n   = 2
        self.noe = 1
        self.noe_bc = 6
        self.dof = 8
        self.basic = sempy.Basic()
        self.D = self.basic.derivative_matrix_gll( self.n )
        self.physical = np.zeros((1),int)
        self.physical = 1
        #self.physical_names, self.theta,,\
        self.points, self.weights = \
                    self.basic.gauss_lobatto_legendre(self.n)
        
        self.x = np.zeros((self.noe, self.n, self.n, self.n),float)
        self.y = np.zeros((self.noe, self.n, self.n, self.n),float)
        self.z = np.zeros((self.noe, self.n, self.n, self.n),float)
        
        self.x[0,0,0,0]=0.3333333333333335
        self.x[0,1,0,0]=0.5000000000013294
        self.x[0,0,1,0]=0.0
        self.x[0,1,1,0]=0.0
        self.x[0,0,0,1]=0.2500000000001332
        self.x[0,1,0,1]=0.3333333333337766
        self.x[0,0,1,1]=0.0
        self.x[0,1,1,1]=0.0
        
        self.y[0,0,0,0]=0.3333333333333335
        self.y[0,1,0,0]=0.4999999999986707
        self.y[0,0,1,0]=0.5000000000013305
        self.y[0,1,1,0]=1.0
        self.y[0,0,0,1]=0.2499999999998673
        self.y[0,1,0,1]=0.3333333333328904
        self.y[0,0,1,1]=0.3333333333333337
        self.y[0,1,1,1]=0.4999999999986717
        
        self.z[0,0,0,0]=0.666666666666667
        self.z[0,1,0,0]=1.0
        self.z[0,0,1,0]=0.5000000000013305
        self.z[0,1,1,0]=1.0
        self.z[0,0,0,1]=0.7500000000001333
        self.z[0,1,0,1]=1.0
        self.z[0,0,1,1]=0.6666666666671102
        self.z[0,1,1,1]=1.0
        
        #self.x[0,0,0,0]=0.0;self.x[0,1,0,0]=1.0;self.x[0,0,1,0]=0.0;self.x[0,1,1,0]=1.0;
        #self.x[0,0,0,1]=0.0;self.x[0,1,0,1]=1.0;self.x[0,0,1,1]=0.0;self.x[0,1,1,1]=1.0;
        #self.y[0,0,0,0]=0.0;self.y[0,1,0,0]=0.0;self.y[0,0,1,0]=1.0;self.y[0,1,1,0]=1.0;
        #self.y[0,0,0,1]=0.0;self.y[0,1,0,1]=0.0;self.y[0,0,1,1]=1.0;self.y[0,1,1,1]=1.0;
        #self.z[0,0,0,0]=0.0;self.z[0,1,0,0]=0.0;self.z[0,0,1,0]=0.0;self.z[0,1,1,0]=0.0;
        #self.z[0,0,0,1]=1.0;self.z[0,1,0,1]=1.0;self.z[0,0,1,1]=1.0;self.z[0,1,1,1]=1.0;
       
        
        self.jac = space_f90.geometric_3d(self.x, self.y, self.z, self.D)
                
        self.theta=np.zeros((self.noe, self.n, self.n, self.n),int)
        
        self.theta[0,0,0,0]=0
        self.theta[0,1,0,0]=1
        self.theta[0,0,1,0]=2
        self.theta[0,1,1,0]=3
        
        self.theta[0,0,0,1]=4
        self.theta[0,1,0,1]=5
        self.theta[0,0,1,1]=6
        self.theta[0,1,1,1]=7
        
       
    #def theta_phys_bc(self,k,m,n):
    #    i    = self.theta_bc[k,m,n]
    #    phys = self.physical_bc[k]
    #    return i, phys
        
    #def theta_phys(self,k,m,n,o):
    #    i    = self.theta[k,m,n,o]
    #    phys = self.physical[k]
    #    return i, phys
        
    def mapping_q(self,v):
        # from global to local
        w = space_f90.mapping_q_3d(v,self.theta)
        return w
        
    def mapping_qt(self,w):
        v = space_f90.mapping_qt_3d(w,self.theta,self.dof)
        return v
        
    def local_to_global(self,w):
        v = space_f90.local_to_global_3d(w,self.theta,self.dof)
        return v
        
    def quadrature(self,f):
        kappa = space_f90.quadrature_3d(self.weights,self.jac,f)
        return kappa
        
           
        
class MeshObject3D():
    '''
    This class creates a sem mesh object based on a gmsh file and the chosen polynomial degree. 
    '''
    def __init__(self,filename,n,n_gmsh=3):
        print ''
        print '**** Creating Space object     ****'
        t1 = time.clock()
        self.GTS      = GmeshToSpectral3D(filename,n_gmsh)
        self.bc_type  = self.GTS.bc_type
        self.dim      = 3
        self.basic = sempy.Basic()
               
        self.physical = self.GTS.physical
        self.n        = n
        self.D        = self.basic.derivative_matrix_gll(self.n)
        self.points, self.weights = \
                    self.basic.gauss_lobatto_legendre(self.n)
        self.noe      = self.GTS.noe
        
        if self.n == n_gmsh:
            self.x = self.GTS.x
            self.y = self.GTS.y
            self.z = self.GTS.z
            self.dof = self.GTS.dof
            self.theta = self.GTS.theta
        else:
            self.x = self.interpolation(self.GTS.x,self.n)
            self.y = self.interpolation(self.GTS.y,self.n)
            self.z = self.interpolation(self.GTS.z,self.n)
            self.theta, self.dof = space_f90.connectivity_3d(self.x,self.y,self.z)
        
        self.jac = space_f90.geometric_3d(self.x, self.y, self.z, self.D)
        #self.jac = self.geometric()
                
        self.theta_bc, self.physical_bc, self.jac_s, self.edge = self.boundary_connectivity()
        self.noe_bc = self.GTS.theta_bc.shape[0]
        #self.n_x, self.n_y = self.surface_unit_normal()
        # Arrays to calculate derivatives
        self.__derivative_geometric()
        
        t2 = time.clock()
        print '**** dim =',self.dim
        print '**** noe =',self.noe
        print '**** dof =',self.dof
        print '****   n =',self.n
        print '**** ex. time=',t2-t1
        print '**** Finished creating instance.'
        
        
    #def plot_mesh(self):
    
    #def plot_scalar(self,u,n_int=0):
               
    #def plot_basis_function(self,n_g):
    def __derivative_geometric(self):
        #
        self.g11 = np.zeros((self.noe,self.n,self.n,self.n),float)
        self.g12 = np.zeros((self.noe,self.n,self.n,self.n),float)
        self.g13 = np.zeros((self.noe,self.n,self.n,self.n),float)
        self.g21 = np.zeros((self.noe,self.n,self.n,self.n),float)
        self.g22 = np.zeros((self.noe,self.n,self.n,self.n),float)
        self.g23 = np.zeros((self.noe,self.n,self.n,self.n),float)
        self.g31 = np.zeros((self.noe,self.n,self.n,self.n),float)
        self.g32 = np.zeros((self.noe,self.n,self.n,self.n),float)
        self.g33 = np.zeros((self.noe,self.n,self.n,self.n),float)
            
        for k in range(self.noe):
            self.g11[k,:,:,:],self.g12[k,:,:,:],self.g13[k,:,:,:],\
            self.g21[k,:,:,:],self.g22[k,:,:,:],self.g23[k,:,:,:],\
            self.g31[k,:,:,:],self.g32[k,:,:,:],self.g33[k,:,:,:] =\
                    operators_f90.geometric_convection_3d(self.x[k,:,:,:],
                                                          self.y[k,:,:,:],
                                                          self.z[k,:,:,:],
                                                          self.D)
    
    def __geometric(self):
        jac   =np.zeros((self.noe, self.n, self.n, self.n),np.float)
        x_xi  =np.zeros((self.noe, self.n, self.n, self.n),np.float)
        x_eta =np.zeros((self.noe, self.n, self.n, self.n),np.float)
        x_zeta=np.zeros((self.noe, self.n, self.n, self.n),np.float)
        y_xi  =np.zeros((self.noe, self.n, self.n, self.n),np.float)
        y_eta =np.zeros((self.noe, self.n, self.n, self.n),np.float)
        y_zeta=np.zeros((self.noe, self.n, self.n, self.n),np.float)
        z_xi  =np.zeros((self.noe, self.n, self.n, self.n),np.float)
        z_eta =np.zeros((self.noe, self.n, self.n, self.n),np.float)
        z_zeta=np.zeros((self.noe, self.n, self.n, self.n),np.float)
        t1    =np.zeros((self.noe, self.n, self.n, self.n),np.float)
        t2    =np.zeros((self.noe, self.n, self.n, self.n),np.float)
        t3    =np.zeros((self.noe, self.n, self.n, self.n),np.float)
        # geometric
        for k in range(self.noe):
          for alpha in range(self.n):
              for beta in range(self.n):
                  for gamma in range(self.n):
                      for i in range(self.n):
                          x_xi[k,alpha,beta,gamma] = x_xi[k,alpha,beta,gamma] + self.D[alpha,i] * self.x[k,i,beta,gamma]
                          y_xi[k,alpha,beta,gamma] = y_xi[k,alpha,beta,gamma] + self.D[alpha,i] * self.y[k,i,beta,gamma]
                          z_xi[k,alpha,beta,gamma] = z_xi[k,alpha,beta,gamma] + self.D[alpha,i] * self.z[k,i,beta,gamma]
        
                          x_eta[k,alpha,beta,gamma] = x_eta[k,alpha,beta,gamma] + self.D[beta,i]* self.x[k,alpha,i,gamma]
                          y_eta[k,alpha,beta,gamma] = y_eta[k,alpha,beta,gamma] + self.D[beta,i]* self.y[k,alpha,i,gamma]
                          z_eta[k,alpha,beta,gamma] = z_eta[k,alpha,beta,gamma] + self.D[beta,i]* self.z[k,alpha,i,gamma]
        
                          x_zeta[k,alpha,beta,gamma] = x_zeta[k,alpha,beta,gamma] + self.D[gamma,i] * self.x[k,alpha,beta,i]
                          y_zeta[k,alpha,beta,gamma] = y_zeta[k,alpha,beta,gamma] + self.D[gamma,i] * self.y[k,alpha,beta,i]
                          z_zeta[k,alpha,beta,gamma] = z_zeta[k,alpha,beta,gamma] + self.D[gamma,i] * self.z[k,alpha,beta,i]
        # Jacobian
        t1[:,:,:,:] = x_xi * ( y_eta * z_zeta - y_zeta * z_eta )
        t2[:,:,:,:] = y_xi * ( x_eta * z_zeta - x_zeta * z_eta )
        t3[:,:,:,:] = z_xi * ( x_eta * y_zeta - x_zeta * y_eta )
        jac[:,:,:,:] = t1 - t2 + t3
        return jac
    
    
    def gradient_vector(self,u):
        # grad
        #print space_f90.gradient_vector_3d.__doc__
        u_x, u_y, u_z = space_f90.gradient_vector_3d(
                                u,
                                self.jac,self.D,
                                self.g11,self.g12,self.g13,
                                self.g21,self.g22,self.g23,
                                self.g31,self.g32,self.g33 )
        return u_x,u_y,u_z
           
    def l2_norm(self,u):
        #u_local = self.mapping_q(u)
        norm = space_f90.l2norm_3d(self.weights,self.jac,u)
        return norm
    
    def quadrature(self,f):
        kappa = space_f90.quadrature_3d(self.weights,self.jac,f)
        #kappa=0.0
        #for k in range(self.noe):
        #    alpha=0.0
        #    for m in range(self.n):
        #        for n in range(self.n):
        #            for o in range(self.n):
        #                alpha = alpha + self.weights[m]*self.weights[n]*self.weights[o]*\
        #                        self.jac[k,m,n,o]*f[k,m,n,o]
        #    #print 'alpha[',k,']=',alpha
        #    kappa=kappa+alpha
        return kappa
                    
    def quadrature_surf(self):
        #kappa = space_f90.quadrature_3d(self.weights,self.jac,f)
        #print space_f90.quadrature_surf_3d.__doc__
        kappa= space_f90.quadrature_surf_3d(self.weights,self.jac_s)
        #kappa = 0.0
        #for k in range(self.noe_bc):
        #    for m in range(self.n):
        #        for n in range(self.n):
        #            #i = self.theta_bc[k,m,n]
        #            #print 'f[',i,']=',f[i]
        #            kappa = kappa + self.weights[m]*self.weights[n]*self.jac_s[k,m,n]#*f[i]
        #            #kappa = kappa + self.jac_s[k,m,n]
                    
        return kappa
        
    def theta_phys(self, k, m, n, o):
        '''Local to global connectivity method. Takes as input the local node 
           description and returns the global node number and the physical name 
           of the element. 

           :param k: Element number
           :param m: Node number in the xi direction
           :param n: Node number in the eta direction
           :type k: integer
           :type m: integer
           :type n: integer
           :returns: An integer tuple of global node number and corresponding physical name.         
           
             
           Example::
              
              for k in range(mesh.noe):
                  for m in range(mesh.n):
                      for n in range(mesh.n):
                          i, phys = mesh.theta_phys(k,m,n)
        '''
        i    = self.theta[k,m,n,o]
        phys = self.physical[k]
        return i, phys

    def theta_phys_bc(self, k, m, n):
        i       = self.theta_bc[k,m,n]
        phys    = self.physical_bc[k]
        boundarycondtion_type = self.bc_type[phys-1]
        element_edge = self.edge[k]
        
        return i, phys, boundarycondtion_type, element_edge
    
    #def gradient_vector(self,u):
    #    u_x,u_y = space_f90.gradient_vector_2d(u,self.x,self.y,self.jac,self.D)
    #    return u_x,u_y
    
    
    #def surface_flux(self,u_grad):
    #    u_bx = self.boundary_value(u_grad[0])
    #    u_by = self.boundary_value(u_grad[1])
    #    #mu_b = self.boundary_value(mu)
    #    
    #    kappa = 0.0
    #    for k in range(self.noe_bc):
    #        for m in range(self.n):
    ##            [i, phys, boundarycondtion_type, element_edge] = self.theta_phys_bc(k,m)
    #             u_n = u_bx[k,m] * self.n_x[k,m] + u_by[k,m] * self.n_y[k,m]
    #             #coeff = (self.n_x[k,m] +self.n_y[k,m])/(self.n_x[k,m]*self.n_y[k,m])
    #             #print coeff
    #             #u_n = u_bx[k,m] * self.n_x[k,m]*self.n_x[k,m] + \
    #             #      u_by[k,m] * self.n_y[k,m]*self.n_y[k,m]
    #             #print u_n
    #             kappa = kappa +self.weights[m]*self.jac_s[k,m]*u_n
    #             
    #    return kappa
    
        
    def surface_unit_normal(self):
        n_x_f, n_y_f = space_f90.surface_1_2d(self.x,self.y,self.D)
        #n_x = np.zeros((self.noe, self.n, self.n),float)
        #n_y = np.zeros((self.noe, self.n, self.n),float)
        n_x = np.zeros((self.noe_bc, self.n),float)
        n_y = np.zeros((self.noe_bc, self.n),float)
        
       
        for k in range(self.noe_bc):
            for m in range(self.n):
                [i, phys, boundarycondtion_type, element_edge] = self.theta_phys_bc(k,m)
                if element_edge[1] == 0:
                    k_e = element_edge[0]
                    #n_x[k_e,m,0]       =  n_x_f[k_e,0,m]
                    #n_y[k_e,m,0]       =  n_y_f[k_e,0,m]
                    n_x[k,m]       =  n_x_f[k_e,0,m]
                    n_y[k,m]       =  n_y_f[k_e,0,m]
                        
                if element_edge[1] == 1:
                    k_e = element_edge[0]
                    #n_x[k_e,m,self.n-1] =  n_x_f[k_e,1,m]
                    #n_y[k_e,m,self.n-1] =  n_y_f[k_e,1,m]
                    n_x[k,m] =  n_x_f[k_e,1,m]
                    n_y[k,m] =  n_y_f[k_e,1,m]
                        
                if element_edge[1] == 2:
                    k_e = element_edge[0]
                    #n_x[k_e,0,m]        =  n_x_f[k_e,2,m]
                    #n_y[k_e,0,m]        =  n_y_f[k_e,2,m]
                    n_x[k,m]        =  n_x_f[k_e,2,m]
                    n_y[k,m]        =  n_y_f[k_e,2,m]
                        
                if element_edge[1] == 3:
                    k_e = element_edge[0]
                    #n_x[k_e,self.n-1,m] = n_x_f[k_e,3,m]
                    #n_y[k_e,self.n-1,m] = n_y_f[k_e,3,m]
                    n_x[k,m] = n_x_f[k_e,3,m]
                    n_y[k,m] = n_y_f[k_e,3,m]
                
        #plt.figure()
        ##k=2
        #for k in range(self.noe):
        #    Q = plt.quiver(self.x[k,:,:],self.y[k,:,:],n_x[k,:,:],n_y[k,:,:],\
        #                   scale=25.0)#,units='dots')#,color='r',linewidths=(2,) )
        ##qk =plt.quiverkey(Q, 0.9, 0.95, 2,'asdf',coordinates='figure')
        ##plt.axis([-0.2,1.2,-0.2,1.2])
        #plt.show()
        return n_x,n_y
                
                
        
    def boundary_value(self,u):
        '''
        Creates an array of boundary values of a function u.
        
        :param u: An array in local data representation
        
        :returns: An array of boundary points.
        '''
        u_bound = np.zeros((self.noe_bc, self.n), float)
        
        for k in range(self.noe_bc):
            for m in range(self.n):
                [i, phys, bc_type, element_edge] = self.theta_phys_bc(k,m)
                k_element = element_edge[0]
                if element_edge[1] == int(0):
                    u_bound[k,m] = u[k_element,m,0]
                if element_edge[1] == int(1):
                    u_bound[k,m] = u[k_element,m,self.n-1]
                if element_edge[1] == int(2):
                    u_bound[k,m] = u[k_element,0,m]
                if element_edge[1] == int(3):
                    u_bound[k,m] = u[k_element,self.n-1,m]
                
        return u_bound
        
    def mapping_q(self,v):
        #  w[k,m,n] = v[i]
        w = space_f90.mapping_q_3d(v,self.theta)
        #w=np.zeros((self.noe, self.n, self.n, self.n),float)
        #for k in range(self.noe):
        #    for m in range(self.n):
        #        for n in range(self.n):
        #            for o in range(self.n):
        #                i = self.theta[k,m,n,o]
        #                w[k,m,n,o]=v[i]
        return w
        
    def mapping_qt(self,w):
        #  v[i] = v[i] + w[k,m,n]
        v = space_f90.mapping_qt_3d(w,self.theta,self.dof)
        #v = np.zeros((self.dof),float)
        #for k in range(self.noe):
        #    for m in range(self.n):
        #        for n in range(self.n):
        #            for o in range(self.n):
        #                i    = self.theta[k,m,n,o]
        #                v[i] = v[i] + w[k,m,n,o]
        return v

    def local_to_global(self,w):
        # v[i]=w[k,m,n]
        v = space_f90.local_to_global_3d(w,self.theta,self.dof)
        #v = np.zeros((self.dof),float)
        #for k in range(self.noe):
        #    for m in range(self.n):
        #        for n in range(self.n):
        #            for o in range(self.n):
        #                i    = self.theta[k,m,n,o]
        #                v[i] = w[k,m,n,o]
        return v
        
    def mask(self,w):
        '''
        This method masks nodal point values associated with a Dirichlet boundary condition.
        
        :param w: Numpy array in global data representation subject to masking.
        :returns: A masked numpy array in global data representation.
        
        '''
        for k in range(self.noe_bc):
            for m in range(self.n):
                for n in range(self.n):
                    [i, name, bc_type, edge] = self.theta_phys_bc(k,m,n)
                    if bc_type == 'Dir':
                        w[i] = 0.0
        return w
        
    def mask_matrix(self,A):
        '''
        This method masks nodal point values associated with a Dirichlet boundary condition.
        
        :param A: Numpy array in global data representation subject to masking.
        :type A: PysparseMatrix
        :returns: A masked PysparseMatrix.
        
        '''
        val = np.zeros((self.dof),float)
        row = np.zeros((self.dof),int)
        col = np.zeros((self.dof),int)
        col[:]=range(self.dof)
        
        for k in range(self.noe_bc):
            for m in range(self.n):
                for n in range(self.n):
                    [i, phys, bc_type, element_edge]=self.theta_phys_bc(k,m,n)
                    if bc_type == "Dir":
                        #for j in range(self.dof):
                        #    A[int(i),j] = 0.0
                        #for j in range(self.dof):
                        #    if np.abs(A[int(i),j])>0.0:
                        #        A[int(i),j] = 0.0
                        row[:] = int(i)
                        A.put(val,row,col)
                        
        return A
        
    def interpolation(self,u,m):
        n_local = u.shape[1]
        points_local, weights_local = \
                    self.basic.gauss_lobatto_legendre( n_local )
        points_int,   weights_int   = \
                    self.basic.gauss_lobatto_legendre( m )
        u_int = space_f90.interpolation_3d( points_int, u, points_local )
        return u_int
        
    def hyper_filter(self,u_tilde,alpha=0.25):
        u1 = self.interpolation(u_tilde,self.n-1)
        u2 = self.interpolation(u1,self.n)
        u_out = alpha * u2 + (1.0 - alpha) * u_tilde
        return u_out
        
    def boundary_connectivity_gmsh(self,j):
        u      = np.zeros((self.GTS.dof),float)
        noe_bc = self.GTS.theta_bc.shape[0]
        
        for k in range(noe_bc):
            for m in range(self.GTS.n):
                for n in range(self.GTS.n):
                    i = self.GTS.theta_bc[k,m,n]
                    if self.GTS.physical_bc[k] == j:
                        u[i] = self.GTS.physical_bc[k]
                    
        return self.GTS.mapping_q(u)
        
         
    def boundary_connectivity_gll(self,u_gmsh):
        #u_gll = np.zeros((self.noe,self.n,self.n),float)
        #for k in range(self.noe):
        #    for i in range(self.n):
        #        for j in range(self.n):
        #            for m in range(self.GTS.n):
        #                for n in range(self.GTS.n):
        #                    L_xi  = basic.lagrange(m,self.GTS.n,self.points[i],self.GTS.points)
        #                    L_eta = basic.lagrange(n,self.GTS.n,self.points[j],self.GTS.points)
        #                    u_gll[k,i,j] = u_gll[k,i,j]+u_gmsh[k,m,n]*L_xi*L_eta
        u_gll = self.interpolation(u_gmsh,self.n)
        return np.rint(u_gll)
        
    def boundary_connectivity_theta(self, theta_bc, physical_bc, jac_s, edge, count, u_gll, i):
    #def boundary_connectivity_theta(self, theta_bc, physical_bc, jac_s, edge, count, u_gll, i):
        #
        for k in range(self.noe):
                alpha1=0.0
                alpha2=0.0
                alpha3=0.0
                alpha4=0.0
                alpha5=0.0
                alpha6=0.0
                for m in range(self.n):
                    for n in range(self.n):
                        alpha1=alpha1 + u_gll[k,m,n,0]
                        alpha2=alpha2 + u_gll[k,m,n,self.n-1]
                        alpha3=alpha3 + u_gll[k,m,0,n]
                        alpha4=alpha4 + u_gll[k,m,self.n-1,n]
                        alpha5=alpha5 + u_gll[k,0,m,n]
                        alpha6=alpha6 + u_gll[k,self.n-1,m,n]
                # zeta planes      
                if int(np.rint(alpha1)) == self.n*self.n*i:
                    for m in range(self.n):
                        for n in range(self.n):
                            theta_bc[count,m,n] = self.theta[k,m,n,0]# edge 0
                    physical_bc[count]=i
                    jac_s[count,:,:] = space_f90.jac_surf_3d_1(self.x[k,:,:,0],\
                                                              self.y[k,:,:,0],\
                                                              self.z[k,:,:,0],self.D)
                    edge[count,0] = k # element number
                    edge[count,1] = 0 # edge number
                    count = count +1
                    
                    
                if int(np.rint(alpha2)) == self.n*self.n*i:
                    for m in range(self.n):
                        for n in range(self.n):
                            theta_bc[count,m,n]=self.theta[k,m,n,self.n-1]# edge 1
                    physical_bc[count]=i
                    jac_s[count,:,:] = space_f90.jac_surf_3d_1(self.x[k,:,:,self.n-1],\
                                                              self.y[k,:,:,self.n-1],\
                                                              self.z[k,:,:,self.n-1],self.D)
                    edge[count,0] = k # element number
                    edge[count,1] = 1 # edge number
                    count = count +1
                     
                # eta planes    
                if int(np.rint(alpha3)) == self.n*self.n*i:
                    for m in range(self.n):
                        for n in range(self.n):
                            theta_bc[count,m,n]=self.theta[k,m,0,n]# edge 2
                    physical_bc[count]=i
                    jac_s[count,:,:] = space_f90.jac_surf_3d_2(self.x[k,:,0,:],\
                                                              self.y[k,:,0,:],\
                                                              self.z[k,:,0,:],self.D)
                    edge[count,0] = k # element number
                    edge[count,1] = 2 # edge number
                    count = count +1
                
                if int(np.rint(alpha4)) == self.n*self.n*i:
                    for m in range(self.n):
                        for n in range(self.n):
                            theta_bc[count,m,n]=self.theta[k,m,self.n-1,n]# edge 3
                    physical_bc[count]=i
                    jac_s[count,:,:] = space_f90.jac_surf_3d_2(self.x[k,:,self.n-1,:],\
                                                              self.y[k,:,self.n-1,:],\
                                                              self.z[k,:,self.n-1,:],self.D)
                    edge[count,0] = k # element number
                    edge[count,1] = 3 # edge number
                    count = count +1
                
                # xi planes
                if int(np.rint(alpha5)) == self.n*self.n*i:
                    for m in range(self.n):
                        for n in range(self.n):
                            theta_bc[count,m,n]=self.theta[k,0,m,n]# edge 4
                    physical_bc[count]=i
                    jac_s[count,:,:] = space_f90.jac_surf_3d_3(self.x[k,0,:,:],\
                                                              self.y[k,0,:,:],\
                                                              self.z[k,0,:,:],self.D)
                    edge[count,0] = k # element number
                    edge[count,1] = 4 # edge number
                    count = count +1
                    
                if int(np.rint(alpha6)) == self.n*self.n*i:
                    for m in range(self.n):
                        for n in range(self.n):
                            theta_bc[count,m,n]=self.theta[k,self.n-1,m,n]# edge 5
                    physical_bc[count]=i
                    jac_s[count,:,:] = space_f90.jac_surf_3d_3(self.x[k,self.n-1,:,:],\
                                                              self.y[k,self.n-1,:,:],\
                                                              self.z[k,self.n-1,:,:],self.D)
                    edge[count,0] = k # element number
                    edge[count,1] = 5 # edge number
                    count = count +1
                    
        return theta_bc, physical_bc, count, jac_s, edge
        #return theta_bc, physical_bc, jac_s, edge, count
        
    def boundary_connectivity(self):
        noe_bc      = self.GTS.theta_bc.shape[0]
        theta_bc    = np.zeros((noe_bc,self.n,self.n), int)
        
        jac_s       = np.zeros((noe_bc,self.n,self.n), float)
        u_gll       = np.zeros((self.noe,self.n,self.n,self.n), float)
        physical_bc = np.zeros((noe_bc), int)
        edge        = np.zeros((noe_bc, 2), int)
        
        #print 'names=',self.GTS.physical_names
        #print 'shape=',self.GTS.physical_names.shape[0]
        
        # Locating boundary points
        #print 'phys'
        #print self.GTS.physical_names
        count=0
        for i in range(self.GTS.physical_names.shape[0]):
            if self.GTS.physical_names[i,0] == 2:
                j = self.GTS.physical_names[i,1]
                u_gmsh = self.boundary_connectivity_gmsh(j)
                u_gll  = self.boundary_connectivity_gll(u_gmsh)
                theta_bc, physical_bc, count, jac_s, edge = \
                     self.boundary_connectivity_theta(theta_bc, physical_bc, jac_s, edge, count, u_gll, j)
            
        return theta_bc, physical_bc, jac_s, edge
        






from basic import *
#import basic_f90


__all__ = ["basic","basic_f90"]

'''
Sempy: A Python implementation of the spectral element method.
Copyright (C) 2011,  Stian Jensen.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
''' 


'''
Sempy: A Python implementation of the spectral element method.
Copyright (C) 2011,  Stian Jensen.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
''' 

import numpy as np
from scipy import linalg
import basic_f90


class Basic():
    '''
    Fundamental SEM routines. 
    
    This class contains methods for computing the GLL and GL 
    quadrature points and weights, the GLL derivative matrix and 
    values of the Legendre interpolants. 
    '''
    def __init__(self):
        pass
    
    def legendre(self,n, xi):
        """
        Computes the value :math:`\kappa` of the :math:`n`'th Legendre polynomial 
        
        .. math::
           \kappa = l_n(\\xi)
           
        evaluated at the point :math:`\\xi`, 
        where :math:`-1\le\\xi\le 1`.
        
        :param n: Number of points. 
        :type n: int
        :param xi: A point.  
        :type xi: float
        :returns: * **kappa** (float) - The value of :math:`l_n(\\xi)`. 
        
        
        **Example:**
        
        >>> import sempy
        >>> basic = sempy.basic.Basic( )
        >>> n = 3
        >>> xi = 0.7
        >>> print basic.legendre( n, xi )
        -0.1925
        """
        if n <= 0: # Just a control structure to avoid hang-up.
            L = 0.0
            print '***: Warning from sempy.basic.legendre: '
            print '---: Input value n was given an invalid value. n=',n
            return L
        
        if n > 0:
            L = basic_f90.legendre(n,xi)
        
        return L
    
    def legendre_derivative(self,n, xi):
        """
        Computes the value of the derivative of the :math:`n`'th Legendre polynomial 
        
        .. math::
           \kappa = l_n'(\\xi)
        
        evaluated at the point :math:`\\xi`, 
        where :math:`-1\le\\xi\le 1`.
                
        :param n: Number of points. 
        :type n: int
        :param xi: A point.  
        :type xi: float
        :returns: * **kappa** (float) - The value of :math:`l_n'(\\xi)`. 
        
        **Example:**
            
        >>> import sempy
        >>> basic = sempy.basic.Basic()
        >>> n = 3
        >>> xi = 0.7
        >>> print basic.legendre_derivative( n, xi )
        2.175
        
        """
        if n <= 0: # Just a control structure to avoid hang-up.
            Ld = 0.0
            print '***: Warning from sempy.basic.legendre_derivative: '
            print '---: Input value n was given an invalid value. n=',n
            return Ld
        if n > 0:
            Ld = basic_f90.legendre_derivative(n, xi)
        
        return Ld
    
    def gauss_legendre(self,n):
        """
        Creates the Gauss-Legendre quadrature points and weights 
        based on the number of points :math:`n`. 
        
        :param n: Number of points. 
        :type n: int
        :returns: * **points** (array) - Quadrature points. 
                  * **weights** (array) - Quadrature weights. 
        
        **Example:**
        
        >>> import sempy
        >>> basic = sempy.basic.Basic()
        >>> n = 3
        >>> points , weights = basic.gauss_legendre( n )
        >>> print points
        [-0.7745966692414834, 1.720588613669501e-17, 0.77459666924148329]
        >>> print weights
        [ 0.55555556  0.88888889  0.55555556]
        
        """
        if n > 1:
            A = basic_f90.gauss_legendre_1( n )
            
            la,v = linalg.eig( A )
            p = sorted( la.real )
    
            # This routine finds the weights
            w = basic_f90.gauss_legendre_2( p )
        else:
            p = 0.0
            w = 0.0
        
        return p,w

    def gauss_lobatto_legendre(self,n):
        """
        Creates the Gauss-Lobatto-Legendre quadrature points and weights 
        based on the number of points :math:`n`. 
        
        :param n: Number of points. 
        :type n: int
        :returns: * **points** (array) - Quadrature points. 
                  * **weights** (array) - Quadrature weights. 
                  
        **Example:**
        
        >>> import sempy
        >>> basic = sempy.basic.Basic()
        >>> n=3
        >>> points , weights = basic.gauss_lobatto_legendre( n )
        >>> print points
        [-1.  0.  1.]
        >>> print weights
        [ 0.33333333  1.33333333  0.33333333]
        """
    
        # Fortran code
        if n < 2:
            p = 0.0
            w = 0.0
            print '***: Warning from sempy.basic.gauss_lobatto_legendre: '
            print '---: Input value n was given an invalid value. n=',n
        
        if n == 2:
            GLpoints = 0.0
            p, w = basic_f90.gauss_lobatto_legendre(GLpoints,n)
                
        if n > 2:
            GLpoints,GLweights = self.gauss_legendre(n-1)
            p, w = basic_f90.gauss_lobatto_legendre(GLpoints,n)
        
        return p,w
    
    def derivative_matrix_gll(self,n):
        """
        This method constructs the derivative matrix 
        of the Lagrange derivatives l'_j(x_i)
        
        Consider the function 
        
        .. math::
        
           u(\\xi)=\sum_{i=0}^{n-1}\ell_i(\\xi)u_i 
        
        Its derivative at the GLL points 
        :math:`\\xi_j` (:math:`0\le j\le n-1`)  
        can be expressed as
        
        .. math::
        
           \\frac{\partial u}{\partial\\xi}\\bigg|_j=
           \sum_{i=0}^{n-1}\ell_i'(\\xi_j)u_i=
           \sum_{i=0}^{n-1}D_{ji}u_i
            
        where :math:`D\in\mathbb R^{n\\times n}` is the derivative matrix. 
           
        :param n: Number of points. 
        :type n: int
        :returns: **D** (array) - Derivative matrix.  
                  
        **Example:**
        
        >>> import sempy
        >>> basic = sempy.basic.Basic()
        >>> n = 3
        >>> D = basic.derivative_matrix_gll( n )
        >>> print D
        [[-1.5  2.  -0.5]
         [-0.5  0.   0.5]
         [ 0.5 -2.   1.5]]

        """
        [points, weights] = self.gauss_lobatto_legendre( n )
        D = basic_f90.lagrange_derivative_matrix( points )
        return D
    
    def lagrange(self,i,xi,gll_points):
        """
        This function returns the value of the :math:`i`'th Lagrangian interpolant 
        evaluated at the point x.
        
        :param i: Index :math:`0\le i\le n-1`. 
        :type i: int
        :param xi: Point.
        :type xi: float
        :param gll_points: GLL quadrature points
                           :math:`\\underline{\\xi}\\in\\mathbb R^{N-1},\\ -1< \\xi_j < 1`. 
        :type gll_points: array
        :returns: Value. 
    
        **Example:**
        
        >>> import sempy
        >>> basic = sempy.basic.Basic()
        >>> n = 3
        >>> p , w = basic.gauss_lobatto_legendre(n)
        >>> xi = 0.3
        >>> print basic.lagrange(n-1, xi, p)
        0.195
        """
        La = basic_f90.lagrange( i+1, xi, gll_points )
        return La

#def legendre(n, x):
#    if n <= 0: # Just a control structure to avoid hang-up.
#        L = 0.0
#        print '***: Warning from sempy.basic.legendre: '
#        print '---: Input value n was given an invalid value. n=',n
#        return L
#     
#    if n > 0:
#        L = basic_f90.legendre(n,x)
#        
#    return L


#def legendre_derivative(n, x):
#    if n <= 0: # Just a control structure to avoid hang-up.
#        Ld = 0.0
#        print '***: Warning from sempy.basic.legendre_derivative: '
#        print '---: Input value n was given an invalid value. n=',n
#        return Ld
#    if n > 0:
#        Ld = basic_f90.legendre_derivative(n, x)
#        
#    return Ld



#def gauss_legendre(n):
#    if n > 1:
#        A = basic_f90.gauss_legendre_1(n)
#    
#        la,v = linalg.eig(A)
#        p = sorted(la.real)
#    
#        # This loop finds the associated weights
#        w = basic_f90.gauss_legendre_2(p)
#    else:
#        p = 0.0
#        w = 0.0
#        
#    return p,w

#def gauss_lobatto_legendre(n):
#    if n < 2:
#        p = 0.0
#        w = 0.0
#        print '***: Warning from sempy.basic.gauss_lobatto_legendre: '
#        print '---: Input value n was given an invalid value. n=',n
#        
#    if n == 2:
#        GLpoints = 0.0
#        p, w = basic_f90.gauss_lobatto_legendre(GLpoints,n)
#                
#    if n > 2:
#        GLpoints,GLweights = gauss_legendre(n-1)
#        p, w = basic_f90.gauss_lobatto_legendre(GLpoints,n)
#        
#    return p,w



#def lagrange_derivative_matrix_gll(n):
#    GLLPoints,GLLWeights = gauss_lobatto_legendre(n)
#    D = basic_f90.lagrange_derivative_matrix(GLLPoints)
#    return D



#def lagrange(i,n,x,gll_points):
#    La = basic_f90.lagrange(i+1,x,gll_points)
#    return La
 

if __name__ == "__main__":
    import doctest
    doctest.testmod()



import sempy
import numpy as np
import matplotlib.pyplot as plt



class NStest:
    '''
    Test case. 
    '''
    def __init__( self, Space, SpaceGL, Time):
        #
        self.Space = Space
        self.SpaceGL = SpaceGL
        self.Time = Time
        self.err_p = sempy.Function( self.SpaceGL )
        self.err_u = sempy.VectorFunction( self.Space )
        
    
    def ic(self,u,p,t):
        #
        x=self.Space.x
        y=self.Space.y
        
        x_gl=self.SpaceGL.x
        y_gl=self.SpaceGL.y
        
        u[0].basis_coeff = -np.cos(x)*np.sin(y)*np.exp(-2.0*t)
        u[1].basis_coeff =  np.sin(x)*np.cos(y)*np.exp(-2.0*t)
        p.basis_coeff = -0.25*(np.cos(2.0*x_gl) + np.cos(2.0*y_gl))*\
                         np.exp(-4.0*t)
        return u
    
    def error_norms( self, u, p, t ):
        # Error norms.
        u_ex = sempy.VectorFunction( self.Space )
        p_ex = sempy.Function( self.SpaceGL )
        
        self.ic( u_ex, p_ex, t )
        #p_ex.plot_wire()
        
        self.err_u[0].basis_coeff = u_ex[0].basis_coeff - u[0].basis_coeff
        self.err_u[1].basis_coeff = u_ex[1].basis_coeff - u[1].basis_coeff
        self.err_p.basis_coeff = p_ex.basis_coeff - p.basis_coeff
        print 'u l2= ',self.err_u[0].l2_norm()
        print 'v l2= ',self.err_u[1].l2_norm()
        print 'p l2= ',self.err_p.l2_norm()
        u_h1 = self.err_u[0].h1_norm()#abs(self.err_u[0].basis_coeff).max()
        v_h1 = self.err_u[1].h1_norm()#abs(self.err_u[1].basis_coeff).max()
        p_h1 = self.err_p.h1_norm()
        print 'u h1= ',u_h1
        print 'v h1= ',v_h1
        print 'p h1= ',p_h1
                
        #return self.err_u[0].l2_norm(), self.err_u[1].l2_norm(),\
        #       self.err_p.l2_norm(),\
        #       u_infty, v_infty, p_infty,self.err_p.h1_norm()
                   
    def bc( self, u, T, t ):
        #
        x=self.Space.x
        y=self.Space.y
        U = -np.cos(x)*np.sin(y)*np.exp(-2.0*t)
        V =  np.sin(x)*np.cos(y)*np.exp(-2.0*t)
        
        for i in range(5):
            u[0].set_bc( U, i )
            u[1].set_bc( V, i )
        
        return
    
  
 


'''
Sempy: A Python implementation of the spectral element method.
Copyright (C) 2011,  Stian Jensen.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
''' 


import numpy as np
import scipy.sparse.linalg as sp
import pysparse
import scipy.linalg as sc

# Compatibility with older versions
if int(pysparse.__version__[2]) >=2:
    from pysparse.sparse.pysparseMatrix import PysparseMatrix #as pysparseMatrix
    
if int(pysparse.__version__[2]) <=1:
    from pysparse.pysparseMatrix import PysparseMatrix
#import sempy.krylov
import time

class CallBack():
    def __init__(self):
        self.iterations = 0
    def it_count(self,u):
        self.iterations=self.iterations + 1
    

class Krylov:
    '''      
    Krylov methods. 
    
    :param pre: Preconditioning matrix, i.e. :math:`P\sim A^{-1}`.
    :type pre: :class:`sempy.precond.Preconditioner`  
    :param float tol: Iterative tolerance. 
    :param integer maxiter: Maximum number of iterations. 
    :param string solver_type: 'cg', 'cgs', 'bicgstab'
    
    :attribute: * **ex_time** (*float*) - Execution time.
    
    :return: The solution vector and a *flag*. 
    
    **Example**::
    
      import sempy
      
      X = sempy.Space( filename = 'square', n = 30, dim = 2 )
      
      laplace = sempy.operators.Laplacian( X )
      A = laplace.matrix

      f = sempy.Function( X, basis_coeff = 1.0 )
      b = sempy.operators.Mass( X ).action_local( f.basis_coeff )
      u = sempy.Function( X, basis_coeff = 0.0 )

      # Solve the system
      solver = sempy.linsolvers.Krylov( solver_type = 'cg' )
      [v, flag] = solver.solve( A, b, u.glob() )
      u.basis_coeff = X.mapping_q(v)
    
    '''
    def __init__(self,pre='none',tol=1e-15,maxiter=5000,solver_type='cg',
                 verbose = 'off' ):
        self.tol         = tol
        self.maxiter     = maxiter
        self.pre         = pre
        self.solver_type = solver_type
        self.cb          = CallBack()
        self.linsolver_type = 'iterative'
        self.verbose = verbose
        self.ex_time =0.0
        self.flag = 0

    def solve(self,A,b,u):
        '''
        Solves the system 
        
        .. math::
        
           Au=b
           
        :param A: System matrix. 
        :param b: Right hand side vector.
        :param u: Solution vector.
        '''
        if self.verbose == 'on':
            print ''
            print '**** Krylov solver            ****'
            print '**** Solver type   =',self.solver_type
        t1 = time.clock()
        self.cb.iterations = 0
        if self.pre == 'none':
            if self.solver_type == 'cg':
                [u, self.flag] = sp.cg( A, b, x0 = u, 
                                        tol = self.tol,
                                        maxiter = self.maxiter,
                                        callback = self.cb.it_count )
            if self.solver_type == 'cgs':
                [u, self.flag] = sp.cgs( A, b, x0 = u, 
                                         tol = self.tol,
                                         maxiter = self.maxiter,
                                         callback = self.cb.it_count )
            if self.solver_type == 'bicgstab':
                [u, self.flag] = sp.bicgstab( A, b, x0 = u, 
                                              tol = self.tol,
                                              maxiter = self.maxiter,
                                              callback = self.cb.it_count )
                
        else:
            if self.solver_type == 'cg':
                [u, self.flag] = sp.cg( A, b, x0 = u, 
                                        tol = self.tol, 
                                        maxiter = self.maxiter,
                                        callback = self.cb.it_count, 
                                        M = self.pre )
                
                #print 'cg res=',np.norm(s)
                #[u, flag] = sempy.krylov.pcg(A,b,u,self.pre,tol=self.tol/10.0,max_iter=self.maxiter)
            if self.solver_type == 'cgs':
                [u, self.flag] = sp.cgs( A, b, x0 = u, 
                                         tol = self.tol,
                                         maxiter = self.maxiter,
                                         callback = self.cb.it_count, 
                                         M = self.pre )
            if self.solver_type == 'bicgstab':
                [u, self.flag] = sp.bicgstab( A, b, x0 = u, 
                                              tol = self.tol,
                                              maxiter = self.maxiter,
                                              callback = self.cb.it_count,
                                              M = self.pre )
        t2 = time.clock()
        #self.ex_time =self.ex_time + (t2-t1)
        self.ex_time =  (t2-t1)
        self.flag = self.cb.iterations
        #s=A*u
        #r=b-s
        #print 'real dot(r)=',np.inner(r,r)
        #flag = flag
        if self.verbose == 'on':
            print '**** Iterations    =',self.flag
            print '**** Solution time =',t2-t1
        
        return u, self.flag
        

        
class Direct:
    '''
    Direct solver. 
    
    :param Space: Discrete function space. 
    :param solver_type: superlu, umfpack. 
    '''
    def __init__(self,Space,solver_type='superlu'):
        self.solver_type = solver_type
        self.Space = Space
        self.linsolver_type = 'direct'
        
    def __convert_to_ll_mat__(self,A):
        Q = pysparse.spmatrix.ll_mat(self.Space.dof, self.Space.dof)
                
        val,irow,jcol = A.find()
        Q.update_add_at(val,irow,jcol)
            
        if self.Space.dim == 1:
            for k in range(self.Space.noe_bc):
                [i, phys, bc_type, element_edge]=self.Space.theta_phys_bc(k)
                if bc_type == "Dir":
                    Q[int(i),int(i)] = 1.0
        if self.Space.dim == 2:
            for k in range(self.Space.noe_bc):
                for m in range(self.Space.n):
                    [i, phys, bc_type, element_edge]=self.Space.theta_phys_bc(k,m)
                    if bc_type == "Dir":
                        Q[int(i),int(i)] = 1.0
        if self.Space.dim == 3:
            for k in range(self.Space.noe_bc):
                for m in range(self.Space.n):
                    for n in range(self.Space.n):
                        [i, phys, bc_type, element_edge]=self.Space.theta_phys_bc(k,m,n)
                        if bc_type == "Dir":
                            Q[int(i),int(i)] = 1.0

        return Q
        
    def solve(self,A,b,u):
        '''
        Solves the system 
        
        .. math::
           Au=b
           
        :param A: System matrix. 
        :param b: Right hand side vector.
        :param u: Solution vector.
        '''
        matrix = self.__convert_to_ll_mat__(A)
        if self.solver_type == 'superlu':    
            LU = pysparse.superlu.factorize(matrix.to_csr())
        if self.solver_type == 'umfpack':
            LU = pysparse.umfpack.factorize(matrix)
        
        # Temporary array
        F_t = np.zeros((self.Space.dof),float)
        F_t[:]=b[:]
        # Add Dirichlet values to the RHS
        if self.Space.dim == 1:
            for k in range(self.Space.noe_bc):
                [i, phys, bc_type, element_edge]=self.Space.theta_phys_bc(k)
                if bc_type == "Dir":
                    F_t[i] = u[i]
        if self.Space.dim == 2:
            for k in range(self.Space.noe_bc):
                for m in range(self.Space.n):
                    [i, phys, bc_type, element_edge]=self.Space.theta_phys_bc(k,m)
                    if bc_type == "Dir":
                        F_t[i] = u[i]
        if self.Space.dim == 3:
            for k in range(self.Space.noe_bc):
                for m in range(self.Space.n):
                    for n in range(self.Space.n):
                        [i, phys, bc_type, element_edge]=self.Space.theta_phys_bc(k,m,n)
                        if bc_type == "Dir":
                            F_t[i] = u[i]
        # Solve            
        LU.solve(F_t,u)
        flag = LU.nnz
        return u, flag
        
        



#import basic

from sempy.space.space import *
from sempy.basic import Basic
from sempy.space.function_vec import *
from sempy.space.plot import *
import operators
import linsolvers
import precond
import fluid
import ode
import solvers
import tests



#from distutils.core import setup
#from distutils.extension import Extension

import ez_setup
ez_setup.use_setuptools()

import os
from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
      name     = 'sempy',
      version = '0.0.18',
      packages = find_packages(),
      include_package_data = True,
      package_data={'': ['*.so','*.f90','*.geo','*.msh']},
      url = 'http://www.sempy.org',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Fortran',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Scientific/Engineering :: Physics'
       ],
      #install_requires = ['numpy>=1.5','scipy>=0.7'],
      #dependency_links = ["https://sourceforge.net/projects/matplotlib/files/matplotlib/matplotlib-1.0.1/"],
      # metadata for upload to PyPI
      author = 'Stian Jensen',
      author_email = 'stianjnsn@gmail.com',
      description = 'A Python implementation of the spectral element method',
      long_description = read("README.rst"),
      license= 'GPL',
      )
